var Util = require('./util'),
    Uuid = require('node-uuid'),
    JsonSchemaValidator = require('json-schema'),
    fs = require('fs'),
    Async = require('async');
    
    

    
    
    
/*

StringTabler interface implemented using a JSON file for local developement.
  
Keep partition in place and replicate all SimpleDB limits. I am only changing
the portions of code that interact with SimpleDb to interact with a local
file instead.
  
*/


var StringTabler = function (inConnectionParams) {

    // inConnectionParams has 'fileName'
    this._filename = inConnectionParams.filename;
    
    // Table definitions -- same as in SimpleDb version so I don't have to change things,
    // except numArributes is always 1
    // Key = Table name in lower case
    // Value = { fieldNameLowerCase: {fieldName, maxByteLength, numAttributes}, ... }
    // There is no __id column here since it is not user-defined and we don't
    // want to allow editing, etc
    this._tables = {};
    
    // Keep simpledb limits in place
    this.POT_BYTES_PER_FIELD = 1024;    // Potential bytes per field with valid UTF-8
    this.GUAR_BYTES_PER_FIELD = 1019;   // Guaranteed bytes per field with valid UTF-8
    
};

// StringTabler interface

StringTabler.prototype.defineTable = function (inTableName, inFieldDefs, inCallback) {

    // Idempotent

    // Validate table name
    // Careful to not allow duplicate table names with different cases
    for (var tableName in this._tables) {
        if (tableName.toLowerCase() === inTableName.toLowerCase()) {
            inCallback({code: 'TABLE_ALREADY_DEFINED'});
            return;
        }
    }
    
    // Validate field definitions {fieldName: maxByteLength, ...}
    var valid = JsonSchemaValidator.validate(
        inFieldDefs,
        {
            type: 'object',
            additionalProperties: {
                type:       'number',
                maxDecimal: 0,
                minimum:    1
            }
        }
    ).valid;
    if (!valid) {
        inCallback({code: 'INVALID_DEFINITION'});
        return;
    }
    
    // Fields cannot contain '__', reserved for system use
    // Couldn't easily check this with JsonSchema regex
    for (var k in inFieldDefs) {
        if (/__/.test(k)) {
            inCallback({code: 'INVALID_FIELD_NAME'});
            return;
        }
    }
    
    // Make sure that field names are unique on a case-insensitive basis
    var k,
        used = {};
    for (k in inFieldDefs) {
        if (used[k.toLowerCase()]) {
            inCallback({code: 'DUPLICATE_FIELD_NAMES'});
            return;
        }
        used[k.toLowerCase()] = true;
    }
    
    // Create the table definition object
    // - Check for duplicate case-insensitive names
    // - Calculate partitioning setup and check max col length
    var fieldName,
        table = {},
        curAttributes,
        totalAttributes = 0;
    for (fieldName in inFieldDefs) {
        
        // Calculate number of attributes that we will have for this field
        curAttributes = Math.ceil(inFieldDefs[fieldName] / this.GUAR_BYTES_PER_FIELD);
        totalAttributes += curAttributes;
        
        // Store the field definition in the table
        table[fieldName] = {
            maxByteLength:  inFieldDefs[fieldName],
            numAttributes:  curAttributes   // Don't want to recalculate this all the time
        };
        
    }
    
    // Give an error if we have more than 256 attributes -- SimpleDb limit
    if (totalAttributes > 256) {
        inCallback({code: 'ROW_TOO_LONG'});
        return;
    }
    
    // Ensure that the table exists in the data
    var data = this._getData();
    if (!data[inTableName]) {
        data[inTableName] = {}; // Indexed by item name
        this._setData(data);
    }
    this._tables[inTableName] = table;
    inCallback();

};

StringTabler.prototype.deleteTable = function (inTableName, inCallback) {

    // Validate the table
    if (!this._tables[inTableName]) {
        inCallback({code: 'NON_EXISTENT_TABLE'});
        return;
    }
   
    // Delete the table
    var data = this._getData();
    delete data[inTableName];
    this._setData(data);
    delete this._tables[inTableName];
    inCallback(null);
    
};


StringTabler.prototype.insertOne = function (inTableName, inRowWithoutId, inCallback) {

    // Convenience function, special case of insertMulti
    this.insertMulti(inTableName, [inRowWithoutId], function (inError, inRowIdArray) {
    
        if (inError) {
            inCallback(inError);
        } else {
            inCallback(null, inRowIdArray[0]);
        }
    });

};

StringTabler.prototype.insertMulti = function (inTableName, inRowArray, inCallback) {

    // Elements of inRowArray may or may not have an __ id property
    // If present, check that it is not already used
    // If missing, assign a new GUID

    // Validate the table and fields
    // Each row may or may not have __id
    
    var err,
        i;
    for (i = 0; i < inRowArray.length; i++) {
        if (inRowArray[i].__id) {
            err = this._validateRowWithId(
                inTableName,
                inRowArray[i],
                false // Do not allow missing fields
            );

        } else {
            err = this._validateRowWithoutId(
                inTableName,
                inRowArray[i],
                false // Do not allow missing fields
            );
        }
        if (err) {
            inCallback({code: err});
            return;
        }
    }

    // Make sure that no duplicate __id values were provided, and
    // calculate the number if item names that we need to generate
    var numNamesNeeded = 0,
        explicitNames = [],
        i,
        curId;
    for (i = 0; i < inRowArray.length; i++) {
        if (curId = inRowArray[i].__id) { // Ignore falsy ids
            if (explicitNames.indexOf(curId) > -1) {
                inCallback({code: 'DUPLICATE_IDS'});
                return;
            } else {
                explicitNames.push(curId);
            }
        } else {
            numNamesNeeded++;
        }
    }

    var generatedNames = [],
        ret,
        me = this;
    Async.series(
        [
            function (cb) {
                
                // Check that all provided __id values are not already present
                if (explicitNames.length === 0) {
                    cb();
                } else {
                    me.selectMulti(
                        inTableName,
                        [], // No fields
                        explicitNames,
                        function (inError, inResults) {

                            
                            if (inError) {
                                cb(inError);
                            } else {

                                // Were there results? inResults is hash remember
                                var res = false;
                                for (var k in inResults) {
                                    res = true;
                                    break;
                                }

                                if (res) {
                                    cb({code: 'ID_EXISTS'});
                                } else {
                                    cb();
                                }
                            
                            }

                            
                        }
                    );
                }

            },
            function (cb) {

                // Create unique item names for those not explicitly assigned
                // Ensure the item names are not in use. This is something to think
                // about, because it doubles the cost and time of insertion, and uuid
                // collision is extremely unlikely. Play it safe for now I suppose.
                Async.whilst(
                    function () {
                        return generatedNames.length < numNamesNeeded;
                    },
                    function (cb) {
                    
                        // Generate however many item names we still need
                        // Make sure there are no duplicates in this list, and
                        // don't duplicate any ids provided explicitly
                        var tryItemNames = [],
                            i,
                            genId;
                        while (tryItemNames.length < numNamesNeeded) {
                            genId = Uuid();
                            if (
                                explicitNames.indexOf(genId) < 0
                                && generatedNames.indexOf(genId) < 0
                                && tryItemNames.indexOf(genId) < 0
                            ) {
                                tryItemNames.push(genId);
                            }
                        }
                        
                        // Check if they exist in the database
                        me.selectMulti(
                            inTableName,
                            [], // No fields
                            tryItemNames,
                            function (inError, inResults) {
                                
                                if (inError) {
                                    cb(inError);
                                } else {

                                    // Items in the results are duplicates
                                    // Remove duplicates from tryItemNames
                                    var k,
                                        idx;
                                    for (k in inResults) {
                                        while ((idx = tryItemNames.indexOf(k)) >= 0) {
                                            tryItemNames.splice(idx, 1);
                                        }
                                    }
                                    
                                    // Copy tryItemNames to generatedNames
                                    for (i = 0; i < tryItemNames.length; i++) {
                                        generatedNames.push(tryItemNames[i]);
                                    }
                                    
                                    cb();
                                    
                                }
                            
                            }
                        );
                    
                    },
                    function (inError) {
                        if (inError) {
                            cb({code: 'ID_GENERATE_ERROR', chain: inError});
                        } else {
                            cb();
                        }
                    }
                );
            },
            function (cb) {
                
                // Write the data
                var putItems = [],
                    i,
                    curItem,
                    curRow,
                    curId,
                    rowWithoutId,
                    data = me._getData();
                ret = [];
                for (var i = 0; i < inRowArray.length; i++) {
                    curRow = inRowArray[i];
                    
                    if (curRow.__id) {
                        curId = inRowArray[i].__id;
                        rowWithoutId = Util.deepCopy(curRow);
                        delete rowWithoutId.__id;
                    } else {
                        curId = generatedNames.pop();
                        rowWithoutId = curRow;
                    }
                    
                    curItem = me._partitionRow(inTableName, rowWithoutId);
                    data[inTableName][curId] = curItem;
                                        
                    ret.push(curId);
                }
                me._setData(data);

                cb();
                
            }
        ],
        function (inError) {

            if (inError) {
                inCallback(inError);
            } else {
                inCallback(null, ret);
            }
            
        }
    );
    
};


StringTabler.prototype.selectOne = function (inTableName, inFieldNames, inRowId, inCallback) {

    // Remember that inFieldNames can be an array or null (all fields)

    // If inFieldNames is not specified then get all fields
    if (arguments.length === 3) {
        inCallback = inFieldNames;
        inFieldNames = [];
    }
       
    
    this.selectMulti(inTableName, inFieldNames, [inRowId], function (inError, inRowWithIdHash) {
    
        if (inError) {
            inCallback(inError);
        } else {
            var ret = inRowWithIdHash[inRowId] || null;
            if (ret) {
                delete ret.__id;
            }
            inCallback(null, ret);
        }
    
        
    });
        

};

StringTabler.prototype.selectMulti = function (inTableName, inFieldNames, inRowIdArray, inCallback) {

    // Remember that inFieldNames can be an array or null (all fields)

    // The only way to get multiple SimpleDb items by name is using
    // a select query with OR expressions (max 20 at a time), unless you want to make one call
    // for each item which seems inefficient
    
    // Validate the table and field names
    var err = this._validateFieldNames(inTableName, inFieldNames);
    if (err) {
        inCallback({code: err});
        return;
    }
    
    // In the special case of asking for an empty array, return here
    // to save resources and avoid errors
    if (inRowIdArray.length === 0) {
        inCallback(null, {});
        return;
    }
    
    // Get the results
    var ret = {},
        i,
        cur,
        add,
        k,
        data = this._getData();
    for (i = 0; i < inRowIdArray.length; i++) {
        cur = data[inTableName][inRowIdArray[i]];
        if (cur) {
    
            // Unpartition
            ret[inRowIdArray[i]] = this._unpartitionRow(inTableName, cur);
            
            // Delete unwanted fields
            if (inFieldNames !== null) {
                for (k in ret[inRowIdArray[i]]) {
                    if (inFieldNames.indexOf(k) < 0) {
                        delete ret[inRowIdArray[i]][k];
                    }
                }
            }
                
            // Add the row id
            ret[inRowIdArray[i]].__id = inRowIdArray[i];
    
        }
    }
    
    inCallback(null, ret);
    
};

StringTabler.prototype.selectQuery = function (inTableName, inFieldNames, inCondition) {

    // Arguments
    var inOrderBy,
        inCallback;
    if (arguments.length === 4) {
        inOptions = {};
        inCallback = arguments[3];
    } else if (arguments.length === 5) {
        inOptions = arguments[3];
        inCallback = arguments[4];
    }

    // Apply default options
    inOptions.orderBy   = inOptions.orderBy || null;
    inOptions.orderDir  = inOptions.orderDir || 'asc';
    inOptions.limit     = inOptions.limit || 0;
    
    // Validate the table and field names (does not allow __id, returned automatically for all queries)
    var err = this._validateFieldNames(inTableName, inFieldNames);
    if (err) {
        inCallback({code: err});
        return;
    }
        
    // Validate condition (does not allow __id, if you had it you would getMulti)
    err = this._validateCondition(inTableName, inCondition);
    if (err) {
        inCallback({code: err});
        return;
    }
    
    // Validate orderField if present
    if (
        inOptions.orderBy
        && !this._tables[inTableName][inOptions.orderBy]
    ) {
        inCallback({code: 'INVALID_ORDER_FIELD'});
        return;
    } else if (
        inOptions.orderBy
        && this._tables[inTableName][inOptions.orderBy].numAttributes > 1
    ) {
        inCallback({code: 'CANNOT_ORDER_ON_PARTITIONED_FIELD'});
        return;
    }   
    
    // Validate orderDirection if present
    if (0 > ['asc', 'desc'].indexOf(inOptions.orderDir)) {
        inCallback({code: 'INVALID_ORDER_DIRECTION'});
        return;
    }
    
    // Get all the rows, unpartition them, add the __id, and check if they match the query
    var data = this._getData(),
        matches = [],
        unpart,
        k,
        i,
        valid;
    for (k in data[inTableName]) {
        unpart = this._unpartitionRow(inTableName, data[inTableName][k]);
        unpart.__id = k;
        
        // Skip rows that don't meet the condition
        valid = true;
        for (i = 0; i < inCondition.length; i++) {
        
            if (
                (inCondition[i][1] === '=' &&      !(unpart[inCondition[i][0]]    === inCondition[i][2]))
                || (inCondition[i][1] === '!=' &&  !(unpart[inCondition[i][0]]    !== inCondition[i][2]))
                || (inCondition[i][1] === '<=' &&  !(unpart[inCondition[i][0]]    <= inCondition[i][2]))
                || (inCondition[i][1] === '>=' &&  !(unpart[inCondition[i][0]]    >= inCondition[i][2]))
                || (inCondition[i][1] === '<' &&   !(unpart[inCondition[i][0]]    < inCondition[i][2]))
                || (inCondition[i][1] === '>' &&   !(unpart[inCondition[i][0]]    > inCondition[i][2]))
            ) {
                valid = false;
                break;
            }
            
        }
        if (valid) {
            matches.push(unpart);
        }
        
    }
   
    // Now we have all the matching rows - sort them
    if (inOptions.orderBy) {
        var fOrderer;
        if (inOptions.orderDir === 'desc') {
            fOrderer = function (a, b) { return (a[inOptions.orderBy] > b[inOptions.orderBy]) ? -1 : 1; };
        } else {
            fOrderer = function (a, b) { return (a[inOptions.orderBy] > b[inOptions.orderBy]) ? 1 : -1; };
        }
        matches.sort(fOrderer);
    }
    
    // Apply the limit
    if (inOptions.limit) { // Do nothing if 0 or null or undefined
        matches = matches.slice(0, inOptions.limit);
    }
    
    // Remove unwanted fields (remember null means all fields)
    var i,
        k;
    if (inFieldNames instanceof Array) {
        for (i = 0; i < matches.length; i++) {
            for (k in matches[i]) {
                if (k !== '__id' && inFieldNames.indexOf(k) < 0) {
                    delete matches[i][k];
                }
            }
        }
    }
    
    inCallback(null, matches);
    
};


StringTabler.prototype.countQuery = function (inTableName, inCondition, inCallback) {

    // Validate condition (does not allow __id, if you had it you would getMulti)
    err = this._validateCondition(inTableName, inCondition);
    if (err) {
        inCallback({code: err});
        return;
    }
    
    // Select the rows and return the count (inefficient, but this is just for dev)
    this.selectQuery(inTableName, [], inCondition, function (inError, inRows) {
        
        if (inError) {
            inCallback(inError);
        } else {
            inCallback(null, inRows.length);
        }
        
    });
    
};


StringTabler.prototype.updateOne = function (inTableName,inRowWithoutId, inRowId, inCallback) {

    // Convenience function, special case of updateMulti
    this.updateMulti(inTableName, inRowWithoutId, [inRowId], function (inError) {
    
        if (inError) {
            inCallback(inError);
        } else {
            inCallback(null);
        }
        
    });

};

StringTabler.prototype.updateMulti = function (inTableName, inRowWithoutId, inRowIdArray, inCallback) {

    // Validate the table and fields
    var err = this._validateRowWithoutId(inTableName, inRowWithoutId, true); // Allow missing fields
    if (err) { 
        inCallback({code: err});
        return;
    }
    
    // Translate the write (i.e. partition)
    var tranRowWithoutId = this._partitionRow(inTableName, inRowWithoutId);
    
    // BatchPutAttributes creates items that don't exist, and using PutAttributes you
    // can only condition the write on the existence of an attribute, not the existence of the item
    // So start by using selectMulti to determine which items actually exist, and write only
    // those -- you lose speed, but this is needed for integrity
    var me = this;
    this.selectMulti(
        inTableName,
        [this._getShortestFieldName(inTableName)],
        inRowIdArray,
        function (inError, inRowWithIdHash) {
                
            if (inError) {
                inCallback({code: 'SELECTMULTI_ERROR', chain: inError});
            } else {
            
                // Create an updated version of the data object
                var data = me._getData(),
                    rowId,
                    attName;
                for (rowId in inRowWithIdHash) { // Only add the ones that exist
                    for (attName in tranRowWithoutId) {
                        data[inTableName][rowId][attName] = tranRowWithoutId[attName];
                    }
                }
                me._setData(data);
            
                inCallback();
                
            }
            
        }
    );
    
};

StringTabler.prototype.updateQuery = function (inTableName, inRowWithoutId, inCondition, inCallback) {

    // Convenience function (with some efficiency gain because you don't need to check
    // for row existence in the query results like updateSingle/Multi do).
    // Get the id of rows that match the query and then run the updates on those rows

    // Validate the table and fields
    var err = this._validateRowWithoutId(inTableName, inRowWithoutId, true); // Allow missing fields
    if (err) { 
        inCallback({code: err});
        return;
    }
        
    // Partition the write (only ince)
    var tranRowWithoutId = this._partitionRow(inTableName, inRowWithoutId);
        
    // The query is validated by selectQuery
    var me = this;
    this.selectQuery(
        inTableName,
        [],
        inCondition, // Validated by selectQuery
        function (inError, inRowWithIdArray) {
                
            if (inError) {
                inCallback(inError); // Intentionally do not chain
            } else {
            
                var rowIds = [],
                    i;
                for (i = 0; i < inRowWithIdArray.length; i++) {
                    rowIds.push(inRowWithIdArray[i].__id);
                }
                me.updateMulti(
                    inTableName,
                    inRowWithoutId,
                    rowIds,
                    inCallback
                );
                
            }
        
        
        }
    );
    
};


StringTabler.prototype.deleteOne = function (inTableName, inRowId, inCallback) {

    // Convenience function, special case of deleteMulti
    this.deleteMulti(inTableName, [inRowId], function (inError) {
    
        if (inError) {
            inCallback(inError);
        } else {
            inCallback(null);
        }
        
    });

};

StringTabler.prototype.deleteMulti = function (inTableName, inRowIdArray, inCallback) {

    // Validate the table
    if (!this._tables[inTableName]) {
        inCallback({'code': 'NON_EXITENT_TABLE'});
        return;
    }
    
    // Remove from data
    var data = this._getData(),
        i;
    for (i = 0; i < inRowIdArray.length; i++) {
        delete data[inTableName][inRowIdArray[i]];
    }
    this._setData(data);
    inCallback();
    
};

StringTabler.prototype.deleteQuery = function (inTableName, inCondition, inCallback) {

    // Convenience function
    // Get the id of rows that match the query and then delete those rows

    // Validate the table
    if (!this._tables[inTableName]) {
        inCallback({'code': 'NON_EXITENT_TABLE'});
        return;
    }
    
    // The query is validated by selectQuery
    var me = this;
    this.selectQuery(
        inTableName,
        [],
        inCondition, // Validated by selectQuery
        function (inError, inRowWithIdArray) {
                        
            if (inError) {
                inCallback(inError); // Intentionally do not chain
            } else {
            
                // Get array of row ids
                var rowIds = [],
                    i;
                for (i = 0; i < inRowWithIdArray.length; i++) {
                    rowIds.push(inRowWithIdArray[i].__id);
                }
            
                // Use deleteMulti to run the API request
                me.deleteMulti(
                    inTableName,
                    rowIds,
                    function (inError) {
                    
                        if (inError) {
                            inCallback(inError); // Purposely not chained
                        } else {
                            inCallback(null);
                        }
                    
                    }
                );
                
            
            }
        
        
        }
    );
    
};


// Private

StringTabler.prototype._partitionRow = function (inTableName, inRow) {

    /*
    
    Translate from client-friendly ({AAA: 'ab', b: 'b'}) to SDB-friendly
    ({AAA_1: 'a', a_2: 'b', ...}). Keep case used by client code, I prevented
    duplicates in defineTable().
    
    Assumes field names and values have already been validated.
    
    You need to return all parts for each field present (i.e. if a field is short you
    still need to write all unused parts to empty strings because we
    might be overwriting them here.
    
    But if a field isn't present at all in inRow we do NOT add it, this would
    break partial row updates.
        
    */
    
    var ret = {},
        k,
        i,
        parts;
        
    for (k in inRow) {
        parts = this._partitionString(inRow[k], 1024); // Split by bytes into UTF-8 valid pieces of at most 1024 bytes
        for (i = 0; i < this._tables[inTableName][k].numAttributes; i++) {
            ret[k + '__' + i] = parts[i] || ''
        }
    }
    
    return ret;

};

StringTabler.prototype._partitionFieldNames = function (inTableName, inFieldNames) {

    /*
    
    Translate from client-friendly ['AAA', 'b'] to SDB-friendly
    ['AAA_0' ,'AAA_1', 'b_0', ...]. Keep case used by client code, I prevented
    duplicates in defineTable().
    Assume field names and values have already been validated.
        
    */
    
    var ret = [],
        i,
        j, 
        curName,
        parts;
        
    for (i = 0; i < inFieldNames.length; i++) {
        curName = inFieldNames[i];
        for (j = 0; j < this._tables[inTableName][curName].numAttributes; j++) {
            ret.push(curName + '__' + j);
        }
    }
    
    return ret;

};

StringTabler.prototype._unpartitionRow = function (inTableName, inTranslatedRow) {

    // Inverse of _partitionRow()
    
    var ret = {},
        k,
        i,
        fieldName,
        splitValue;
        
    // Recomibine fields
    for (k in inTranslatedRow) {
        splitValue = k.split('__');
        fieldName = splitValue[0];
        
        if (!ret[fieldName]) {
            
            i = 0;
            ret[fieldName] = '';
            while (inTranslatedRow[fieldName + '__' + i]) {
                ret[fieldName] += inTranslatedRow[fieldName + '__' + i];
                i++;
            }
            
        }
    
    }
    
    return ret;
    
};

StringTabler.prototype._getShortestFieldName = function (inTableName) {

    // Return the shortest field in the table
    var smMaxByteLength = Infinity,
        smFieldName,
        fieldName,
        curField;
    for (fieldName in this._tables[inTableName]) {
        curField = this._tables[inTableName][fieldName];
        if (curField.maxByteLength < smMaxByteLength) {
            smMaxByteLength = curField.maxByteLength;
            smFieldName = fieldName;
        }
    }
    
    return smFieldName;
    
};

// Validators: return string error code or '' for success
// Assume the object structure is correct (it probably would be costly
// to verify every single object again the schema)


StringTabler.prototype._validateRowWithoutId = function (inTableName, inRowWithoutId, inAllowMissingFields) {

    var table,
        fieldName;
    
    // Verify the table
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Verify that the fields and values are consistent with the table definition.
    for (fieldName in table) { // fieldName key will be lower-case - good
        if (inRowWithoutId[fieldName] === undefined) {
            if (!inAllowMissingFields) {
                return 'MISSING_FIELD';
            }
        } else if (typeof inRowWithoutId[fieldName] !== 'string') {
            return 'VALUE_NOT_STRING';
        } else if (Util.stringByteLength(inRowWithoutId[fieldName]) > table[fieldName].maxByteLength) {
            return 'VALUE_TOO_LONG';
        }
    }
    for (fieldName in inRowWithoutId) {
        if (!this._tables[inTableName][fieldName]) {
            return 'EXTRANEOUS_FIELD';
        }
    }
    
    return '';

};

StringTabler.prototype._validateRowWithId = function (inTableName, inRowWithId, inAllowMissingFields) {

    var table,
        fieldName;
    
    // Verify the table
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Verify that the fields and values are consistent with the table definition.
    for (fieldName in table) { // fieldName key will be lower-case - good
        if (inRowWithId[fieldName] === undefined) {
            if (!inAllowMissingFields) {
                return 'MISSING_FIELD';
            }
        } else if (typeof inRowWithId[fieldName] !== 'string') {
            return 'VALUE_NOT_STRING';
        } else if (Util.stringByteLength(inRowWithId[fieldName]) > table[fieldName].maxByteLength) {
            return 'VALUE_TOO_LONG';
        }
    }
    for (fieldName in inRowWithId) {
        if (!this._tables[inTableName][fieldName] && fieldName !== '__id') {
            return 'EXTRANEOUS_FIELD';
        }
    }
    
    return '';

};

StringTabler.prototype._validateRowsWithoutIdHash = function (inTableName, inRowsWithoutIdHash, inAllowMissingFields) {

    var err,
        k;
    for (k in inRowsWithoutIdHash) {
        err = this._validateRowWithoutId(
            inTableName,
            inRowsWithoutIdHash[k],
            inAllowMissingFields
        );
        if (err) {
            return err;
        }
    }
    
    return '';
    
};

StringTabler.prototype._validateFieldNames = function (inTableName, inFieldNames) {

    // Field names can be empty (no fields) or null (all fields)
    // Make sure that the table exists, and that it has all these field names
    // Missing field names are always allowable
    // inAllowId indicates where the '__id' field is allowed
    
    var i,
        table,
        fieldName;
        
    // Verify the type
    if (!(inFieldNames instanceof Array) && inFieldNames !== null) {
        return 'INVALID_FIELD_LIST';
    }
    
    // Verify the table
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Verify that the fields and values are consistent with the table definition
    if (inFieldNames instanceof Array) {
        for (i = 0; i < inFieldNames.length; i++) {
            fieldName = inFieldNames[i];
            if (!this._tables[inTableName][fieldName]) {
                return 'NON_EXISTENT_FIELD';
            }
        }
    }
    
    return '';

};

StringTabler.prototype._validateCondition = function (inTableName, inCondition) {

    var table,
        condition,
        condFieldName,
        condOperator,
        condValue,
        i;
        
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Check the structure
    // Want to catch sketchily-defined conditions that result
    // in unintended writes
    if (!(inCondition instanceof Array)) {
        return 'INVALID_CONDITION';
    } else {
        for (i = 0; i < inCondition.length; i++) {
            if (!(inCondition[i] instanceof Array) || inCondition[i].length != 3) {
                return 'INVALID_CONDITION';
            }
        }
    }    
    
    // Check each condition
    for (i = 0; i < inCondition.length; i++) {
    
        condFieldName   = inCondition[i][0];
        condOperator    = inCondition[i][1];
        condValue       = inCondition[i][2];
        
        if (!table[condFieldName]) {
            return 'NON_EXISTENT_FIELD';
        } else if (table[condFieldName].numAttributes > 1) {
            return 'CANNOT_CONDITION_ON_PARTITIONED_FIELD';
        } else if (typeof condValue !== 'string') {
            return 'VALUE_NOT_STRING';
        } else if (Util.stringByteLength(condValue) > table[condFieldName].maxByteLength) {
            return 'VALUE_TOO_LONG';
        } else if (0 > ['=', '!=', '<', '>', '<=', '>='].indexOf(condOperator)) {
            return 'INVALID_OPERATOR';
        }
    
    }    
    
    return '';

};


// String functions don't really belong in this class, but oh well

StringTabler.prototype._partitionString = function (inString, inMaxBytesPerPart) {

    // This is used to break UTF-8 strings into partitions  with the
    // byte length of each part not exceeding inMaxBytesPerPart
    // This is because simpledb attributes have a maximum byte length, which
    // is different than characters for UTF-8
    
    var stringLeft = inString,
        buffer = new Buffer(inMaxBytesPerPart),
        bytesWritten,
        chunk,
        chunks = [];
        
    while (stringLeft.length > 0) {
    
        bytesWritten = buffer.write(stringLeft, 'utf8'); // Will not write partial characters if too long (see doc)
        chunks.push(chunk = buffer.toString('utf8', 0, bytesWritten));
        stringLeft = stringLeft.substr(chunk.length);
        
    }
    
    return chunks;
    
};


// Get and set the JSON file data

StringTabler.prototype._getData = function () {

    // Read in the data if the file exists
    // key = table name, value = hash of 'items', key = item name, value = object
    try {
        return JSON.parse(fs.readFileSync(this._filename));
    } catch (e) {
        return {};
    }
    
};

StringTabler.prototype._setData = function (inData) {

    fs.writeFileSync(
        this._filename,
        JSON.stringify(inData),
        'utf8'
    );

};









module.exports = StringTabler;
