var Util = require('./util'),
    Uuid = require('node-uuid'),
    JsonSchemaValidator = require('json-schema'),
    SimpleDb = require('simpledb'),
    Async = require('async');
    
    
    
    
    

    
    
    
/*

StringTabler interface implemented using SimpleDB.

Enforces a tabular data structure of UTF-8 strings. Puts no constraints on the
strings except for length (in bytes, not characters). Accepts any valid UTF-8
string including non-printable characters. No special values like nulls.

- You cannot use fields longer than 1024 bytes in conditions or sorts (since in
my implementation these are partitioned across multiple attributes).

ONLY DIFFERENCE FROM TABLER INTERFACE:

inFieldDefs = {
    fieldName: maxByteLength,
    ...
}

IMPLEMENTATION DETAILS:

- All fields are written to SDB with attribute name fieldName__#, even if partitioning isn't required

- Uses consistent reads (this is the default).

- I use one SimpleDb domain per table. If it doesn't exist,
I create it on defineTable. If it does, I assume it is compatible with the field definitions
send to defineTable.
  
- For each read, I read all field partitions, and for each write, I write all
field partitions (setting unneeded ones to ''). This uses up a bit of extra space in SDB,
but I don't want to mess with deleting unneeded attributes when I update a row.
And you would always need to read them all anyway, since you don't know how long
that particular value will be.
  
- Atomicity is guaranteed under if the write affects 25 rows or fewer, but cannot be
guaranteed beyong that (i.e. the call may fail and write some data). There is
no way around this.
  
  
HOW SIMPLEDB LIMITS ARE APPROACHED

Limitation: Domain size, 10 GB per domain, 1 billion attributes per domain
Approach:   Monitor and redesign app (down the road maybe have transparent table sharding)

Limitation: Domain name 3-255 characters (a-z, A-Z, 0-9, '_', '-', and '.')
Apprach:    Hard limit on table names and domain namespaces

Limitation: Domains per account = 250
Approach:   Request more (this is apparently easy)
 
Limitation: Max Attribute name-value pairs per item = 256
Approach:   Hard limit on field defs (this should be lots for non-text uses)

Limitation: Max attribute name length = 1024 bytes
Approach:   Hard limit on field names

Limitation: Max attribute value length = 1024 bytes
Approach:   Field partitioning

Limitation: Max item name length = 1024 bytes
Approach:   I use GUIDs  no problem

Limitation: Max attributes per PutAttributes operation = 256
Approach:   This is all the attributes for the item  no problem

Limitation: Max items per BatchPutAttributes operation = 25
Approach:   Parallel requests of there are more than 25 items

Limitation: Max attributes requested per Select operation = 256
Approach:   This is the most attributes per item, and all items have the same attributes, so will never run into this barrier.


Limitation: Maximum items in Select response = 2500
Approach:   Iterate on nextToken

Limitation: Maximum query execution time = 5s
Approach:   Iterate on nextToken

Limitation: Maximum number of unique attributes per Select expression = 20
Approach:   The select expression is the where clause -- can only refer to a maximum of 20 unique attributes. Hard limit

Limitation: Maximum number of comparisons per Select expression = 20
Approach:   Hard limit

Limitation: Maximum response size for Select = 1mb
Approach:   Iterate on nextToken
  
*/


var StringTabler = function (inConnectionParams) {

    /*
    
    Connection parameters are store-specific.
    
    inConnectionParams = {
        accessKeyId
        secretAccessKey
        domainNameSpace
    }
    
    */
    
    // Properties
    this._sdb = new SimpleDb.SimpleDB({ // Defaults to consistent reads
        keyid:      inConnectionParams.accessKeyId,
        secret:     inConnectionParams.secretAccessKey
    });
    this._domainNamespace = inConnectionParams.domainNamespace;
    
    // Table definitions
    // Key = Table name
    // Value = { fieldName: {maxByteLength, numAttributes}, ... }
    this._tables = {};
    
    // SDB attributes can hold a maximum of 1024 bytes, but actually 
    // each attribute is only guaranteed to hold 1019 bytes (in the
    // case of a six-byte UTF-8 character getting pushed to the next
    // attribute to keep this attribute valid UTF-8 we have five unused byes)
    this.POT_BYTES_PER_FIELD = 1024;    // Potential bytes per field with valid UTF-8
    this.GUAR_BYTES_PER_FIELD = 1019;   // Guaranteed bytes per field with valid UTF-8
    
  
};

// StringTabler interface

StringTabler.prototype.defineTable = function (inTableName, inFieldDefs, inCallback) {

    // Idempotent

    // Validate table name
    // Careful to not allow duplicate table names with different cases
    for (var tableName in this._tables) {
        if (tableName.toLowerCase() === inTableName.toLowerCase()) {
            inCallback({code: 'TABLE_ALREADY_DEFINED'});
            return;
        }
    }
    
    // Validate field definitions {fieldName: maxByteLength, ...}
    var valid = JsonSchemaValidator.validate(
        inFieldDefs,
        {
            type: 'object',
            additionalProperties: {
                type:       'number',
                maxDecimal: 0,
                minimum:    1
            }
        }
    ).valid;
    if (!valid) {
        inCallback({code: 'INVALID_DEFINITION'});
        return;
    }
    
    // Fields cannot contain '__', reserved for system use
    // Couldn't easily check this with JsonSchema regex
    for (var k in inFieldDefs) {
        if (/__/.test(k)) {
            inCallback({code: 'INVALID_FIELD_NAME'});
            return;
        }
    }
    
    // Make sure that field names are unique on a case-insensitive basis
    var k,
        used = {};
    for (k in inFieldDefs) {
        if (used[k.toLowerCase()]) {
            inCallback({code: 'DUPLICATE_FIELD_NAMES'});
            return;
        }
        used[k.toLowerCase()] = true;
    }
    
    // Create the table definition object
    // - Check for duplicate case-insensitive names
    // - Calculate partitioning setup and check max col length
    var fieldName,
        table = {},
        curAttributes,
        totalAttributes = 0;
    for (fieldName in inFieldDefs) {
        
        // Calculate number of attributes that we will have for this field
        curAttributes = Math.ceil(inFieldDefs[fieldName] / this.GUAR_BYTES_PER_FIELD);
        totalAttributes += curAttributes;
        
        // Store the field definition in the table
        table[fieldName] = {
            maxByteLength:  inFieldDefs[fieldName],
            numAttributes:  curAttributes   // Don't want to recalculate this all the time
        };
        
    }
    
    // Give an error if we have more than 256 attributes -- SimpleDb limit
    if (totalAttributes > 256) {
        inCallback({code: 'ROW_TOO_LONG'});
        return;
    }
    
    // Ensure that the domain exists
    var me = this;
    Async.series(
        [
            function (cb) {
                
                me._sdb.listDomains(function (inError, inDomains) {
                
                    if (inError) {
                        cb({code: 'SIMPLEDB_LISTDOMAINS_ERROR', chain: inError});
                    } else {
                        
                        if (inDomains.indexOf(me._getDomainName(inTableName)) < 0) {
                        
                            me._sdb.createDomain(
                                me._getDomainName(inTableName),
                                function (inError) {
                                    
                                    if (inError) {
                                        cb({code: 'SIMPLEDB_CREATEDOMAIN_ERROR', chain: inError});
                                    } else {
                                        cb();
                                    }
                                }
                            );
                        
                        } else {
                            cb();
                        }
                        
                    }
                        
                });
            
            }
        ],
        function (inError) {
            
            // Save the table defition if it worked and then return
            if (!inError) {
                me._tables[inTableName] = table;
            }
            inCallback(inError);
            
        }
    );

};

StringTabler.prototype.deleteTable = function (inTableName, inCallback) {

    // Validate the table
    if (!this._tables[inTableName]) {
        inCallback({code: 'NON_EXISTENT_TABLE'});
        return;
    }
    
    // Delete the table
    var me = this;
    me._sdb.deleteDomain(
        this._getDomainName(inTableName),
        function (inError) {
            
            if (inError) {
                inCallback({code: 'SIMPLEDB_DELETEDOMAIN_ERROR', chain: inError});
            } else {
                delete me._tables[inTableName];
                inCallback(null);
            }
            
        }
    );
    
};


StringTabler.prototype.insertOne = function (inTableName, inRowWithoutId, inCallback) {

    // Convenience function, special case of insertMulti
    this.insertMulti(inTableName, [inRowWithoutId], function (inError, inRowIdArray) {
    
        if (inError) {
            inCallback(inError);
        } else {
            inCallback(null, inRowIdArray[0]);
        }
    });

};

StringTabler.prototype.insertMulti = function (inTableName, inRowArray, inCallback) {

    // Elements of inRowArray may or may not have an __ id property
    // If present, check that it is not already used
    // If missing, assign a new GUID

    // Validate the table and fields
    // Each row may or may not have __id
    
    var err,
        i;
    for (i = 0; i < inRowArray.length; i++) {
        if (inRowArray[i].__id) {
            err = this._validateRowWithId(
                inTableName,
                inRowArray[i],
                false // Do not allow missing fields
            );

        } else {
            err = this._validateRowWithoutId(
                inTableName,
                inRowArray[i],
                false // Do not allow missing fields
            );
        }
        if (err) {
            inCallback({code: err});
            return;
        }
    }

    // Make sure that no duplicate __id values were provided, and
    // calculate the number if item names that we need to generate
    var numNamesNeeded = 0,
        explicitNames = [],
        i,
        curId;
    for (i = 0; i < inRowArray.length; i++) {
        if (curId = inRowArray[i].__id) { // Ignore falsy ids
            if (explicitNames.indexOf(curId) > -1) {
                inCallback({code: 'DUPLICATE_IDS'});
                return;
            } else {
                explicitNames.push(curId);
            }
        } else {
            numNamesNeeded++;
        }
    }

    var generatedNames = [],
        ret,
        me = this;
    Async.series(
        [
            function (cb) {
                
                // Check that all provided __id values are not already present
                if (explicitNames.length === 0) {
                    cb();
                } else {
                    me.selectMulti(
                        inTableName,
                        [], // No fields
                        explicitNames,
                        function (inError, inResults) {

                            
                            if (inError) {
                                cb(inError);
                            } else {

                                // Were there results? inResults is hash remember
                                var res = false;
                                for (var k in inResults) {
                                    res = true;
                                    break;
                                }

                                if (res) {
                                    cb({code: 'ID_EXISTS'});
                                } else {
                                    cb();
                                }
                            
                            }

                            
                        }
                    );
                }

            },
            function (cb) {

                // Create unique item names for those not explicitly assigned
                // Ensure the item names are not in use. This is something to think
                // about, because it doubles the cost and time of insertion, and uuid
                // collision is extremely unlikely. Play it safe for now I suppose.
                Async.whilst(
                    function () {
                        return generatedNames.length < numNamesNeeded;
                    },
                    function (cb) {
                    
                        // Generate however many item names we still need
                        // Make sure there are no duplicates in this list, and
                        // don't duplicate any ids provided explicitly
                        var tryItemNames = [],
                            i,
                            genId;
                        while (tryItemNames.length < numNamesNeeded) {
                            genId = Uuid();
                            if (
                                explicitNames.indexOf(genId) < 0
                                && generatedNames.indexOf(genId) < 0
                                && tryItemNames.indexOf(genId) < 0
                            ) {
                                tryItemNames.push(genId);
                            }
                        }
                        
                        // Check if they exist in the database
                        me.selectMulti(
                            inTableName,
                            [], // No fields
                            tryItemNames,
                            function (inError, inResults) {
                                
                                if (inError) {
                                    cb(inError);
                                } else {

                                    // Items in the results are duplicates
                                    // Remove duplicates from tryItemNames
                                    var k,
                                        idx;
                                    for (k in inResults) {
                                        while ((idx = tryItemNames.indexOf(k)) >= 0) {
                                            tryItemNames.splice(idx, 1);
                                        }
                                    }
                                    
                                    // Copy tryItemNames to generatedNames
                                    for (i = 0; i < tryItemNames.length; i++) {
                                        generatedNames.push(tryItemNames[i]);
                                    }
                                    
                                    cb();
                                    
                                }
                            
                            }
                        );
                    
                    },
                    function (inError) {
                        if (inError) {
                            cb({code: 'ID_GENERATE_ERROR', chain: inError});
                        } else {
                            cb();
                        }
                    }
                );
            },
            function (cb) {
                
                // Create the argument for the SDB call, and the output
                // for the callback at the same time (ids need to be ordered
                // in the output)
                var putItems = [],
                    i,
                    curItem,
                    curRow,
                    curId,
                    rowWithoutId;
                ret = [];
                for (var i = 0; i < inRowArray.length; i++) {
                    curRow = inRowArray[i];
                    if (curRow.__id) {
                        curId = inRowArray[i].__id;
                        rowWithoutId = Util.deepCopy(curRow);
                        delete rowWithoutId.__id;
                    } else {
                        curId = generatedNames.pop();
                        rowWithoutId = curRow;
                    }
                    curItem = me._partitionRow(inTableName, rowWithoutId);
                    curItem['$ItemName'] = curId;
                    putItems.push(curItem);
                    ret.push(curId);
                }
                
                // SDB call (wrapper takes care of 25-item batching required by SDB)
                me._wrapBatchPutItem(
                    me._getDomainName(inTableName),
                    putItems,
                    function (inError) {
                    
                        if (inError) {
                            cb({code: 'SIMPLEDB_BATCHPUTITEMS_ERROR', chain: inError});
                        } else {
                            cb();
                        }
                    
                    }
                );

            }
        ],
        function (inError) {

            if (inError) {
                inCallback(inError);
            } else {
                inCallback(null, ret);
            }
            
        }
    );
    
};


StringTabler.prototype.selectOne = function (inTableName, inFieldNames, inRowId, inCallback) {

    // Remember that inFieldNames can be an array or null (all fields)

    // If inFieldNames is not specified then get all fields
    if (arguments.length === 3) {
        inCallback = inFieldNames;
        inFieldNames = [];
    }
    
    // Validate the table and field names
    var err = this._validateFieldNames(inTableName, inFieldNames);
    if (err) {
        inCallback({code: err});
        return;
    }     
    
    // The Node.js SimpleDB module doesn't have a way to directly
    // specify particular attributes, need to use the override feature
    // If no field names specified ({} override), will get all attributes
    // Need to specify the names of the field parts
    var tranFields,
        override = {},
        i;
        
    if (inFieldNames instanceof Array && inFieldNames.length === 0) {
        // Note that the item is not returned if it has none of the
        // specified attributes. So ask for something small and don't
        // return it to the called
        override['AttributeName.0'] = this._getShortestFieldName(inTableName) + '__0';
    } else if (inFieldNames instanceof Array) {
        tranFields = this._partitionFieldNames(inTableName, inFieldNames);
        for (i = 0; i < tranFields.length; i++) {
            override['AttributeName.' + i] = tranFields[i];
        }
    }
    
    // Use getAttributes for simple queries, faster than select by 2.4x
    var me = this;
    this._sdb.getItem(
        this._getDomainName(inTableName),
        inRowId,
        override,
        function (inError, inResult) {
        
            if (inError) {
                inCallback({code: 'SIMPLEDB_GETITEM_ERROR', chain: inError});
            } else if (inResult === null) { // Not found
                inCallback(null, null);
            } else {
            
                // Join the field parts back together (if applicable) and return
                var ret;
                if (inFieldNames instanceof Array && inFieldNames.length === 0) {
                    ret = {};
                } else {
                    delete inResult['$ItemName'];
                    ret = me._unpartitionRow(inTableName, inResult);
                }
                inCallback(null, ret);
                
            }
            
        }
    );

};

StringTabler.prototype.selectMulti = function (inTableName, inFieldNames, inRowIdArray, inCallback) {

    // Remember that inFieldNames can be an array or null (all fields)

    // The only way to get multiple SimpleDb items by name is using
    // a select query with OR expressions (max 20 at a time), unless you want to make one call
    // for each item which seems inefficient
    
    // Validate the table and field names
    var err = this._validateFieldNames(inTableName, inFieldNames);
    if (err) {
        inCallback({code: err});
        return;
    }
    
    // In the special case of asking for an empty array, return here
    // to save resources and avoid errors
    if (inRowIdArray.length === 0) {
        inCallback(null, {});
        return;
    }
    
    // Remember that you can only have 20 unique items in your SELECT
    // predicates per query, so you need to batch this
    var slices = [];
    for (i = 0; i < inRowIdArray.length; i += 20) {
        slices.push(inRowIdArray.slice(i, i + 20));
    }    
    
    // Run the parallel requests
    var me = this,
        ret = {};
    Async.forEach( // Parallel
        slices,
        function (inSlice, cb) {
            
            // Build the select statement (there is at least one ID)
            var statement = 'SELECT ',
                tranFields,
                i;
            if (inFieldNames instanceof Array) {
                if (inFieldNames.length === 0) {
                    statement += 'itemName()';
                } else {
                    tranFields = me._partitionFieldNames(inTableName, inFieldNames);
                    for (i = 0; i < tranFields.length; i++) {
                        statement += me._escapeName(tranFields[i]);
                        if (i + 1 < tranFields.length) {
                            statement += ',';
                        }
                    }
                }
            } else if (inFieldNames === null) {
                statement += '*';
            }
            statement += ' FROM ' + me._escapeName(me._getDomainName(inTableName)) + ' WHERE ';
            for (i = 0; i < inSlice.length; i++) {
                if (i > 0) {
                    statement += ' OR ';
                }
                statement += 'itemName() = ' + me._escapeValue(inSlice[i]);
            }
                
            // Run the query
            // Wrapper takes care of NextTokens as appropriate so that all results are returned
            me._wrapSelect(
                statement,
                function (inError, inResult) {
                
                    if (inError) {
                        cb({code: 'SIMPLEDB_SELECT_ERROR', chain: inError});
                    } else {
                        
                        // Join the field parts back together and return
                        // This is safe even if no fields were selected
                        var itemName;
                        for (var i = 0; i < inResult.length; i++) {
                            itemName = inResult[i]['$ItemName'];
                            delete inResult[i]['$ItemName'];
                            ret[itemName] = me._unpartitionRow(inTableName, inResult[i]);
                            ret[itemName]['__id'] = itemName;
                        }         
                        
                        cb(null);
                        
                    }
                    
                }
            );
            
        }, function (inError) {
            
            if (inError) {
                inCallback(inError); // Pre-chained
            } else {
                inCallback(null, ret);
            }
        
        }
    );
    
    
};

StringTabler.prototype.selectQuery = function (inTableName, inFieldNames, inCondition) {

    // Arguments
    var inOrderBy,
        inCallback;
    if (arguments.length === 4) {
        inOptions = {};
        inCallback = arguments[3];
    } else if (arguments.length === 5) {
        inOptions = arguments[3];
        inCallback = arguments[4];
    }

    // Apply default options
    inOptions.orderBy   = inOptions.orderBy || null;
    inOptions.orderDir  = inOptions.orderDir || 'asc';
    inOptions.limit     = inOptions.limit || 0;
    
    // Validate the table and field names (does not allow __id, returned automatically for all queries)
    var err = this._validateFieldNames(inTableName, inFieldNames);
    if (err) {
        inCallback({code: err});
        return;
    }
        
    // Validate condition (does not allow __id, if you had it you would getMulti)
    err = this._validateCondition(inTableName, inCondition);
    if (err) {
        inCallback({code: err});
        return;
    }
    
    // Validate orderField if present
    if (
        inOptions.orderBy
        && !this._tables[inTableName][inOptions.orderBy]
    ) {
        inCallback({code: 'INVALID_ORDER_FIELD'});
        return;
    } else if (
        inOptions.orderBy
        && this._tables[inTableName][inOptions.orderBy].numAttributes > 1
    ) {
        inCallback({code: 'CANNOT_ORDER_ON_PARTITIONED_FIELD'});
        return;
    }   
    
    // Validate orderDirection if present
    if (0 > ['asc', 'desc'].indexOf(inOptions.orderDir)) {
        inCallback({code: 'INVALID_ORDER_DIRECTION'});
        return;
    }
    
    // SimpleDB will not let you order by an attribute unless it is included
    // in a predicate in the WHERE clause. So if the user hasn't already done
    // this, then include a definitely-satisfied condition on that attribute
    if (inOptions.orderBy) {
        var inWhere = false;
        for (var i = 0; i < inCondition.length; i++) {
            if (inCondition[i][0] === inOptions.orderBy) {
                inWhere = true;
                break;
            }
        }
        if (!inWhere) {            
            inCondition.push([inOptions.orderBy, '>=', "\u0000"]);
        }
    }
    
    
    // Build the select statement
    var statement = 'SELECT ',
        tranFields,
        i,
        curCondition;
    if (inFieldNames instanceof Array) {
        if (inFieldNames.length === 0) {
            statement += 'itemName()';
        } else {
            tranFields = this._partitionFieldNames(inTableName, inFieldNames);
            for (i = 0; i < tranFields.length; i++) {
                statement += this._escapeName(tranFields[i]);
                if (i + 1 < tranFields.length) {
                    statement += ',';
                }
            }
        }
    } else if (inFieldNames === null) {
        statement += '*';
    }
    statement += ' FROM ' + this._escapeName(this._getDomainName(inTableName));
    for (i = 0; i < inCondition.length; i++) {
        curCondition = inCondition[i];
        if (i === 0) {
            statement += ' WHERE ';
        } else {
            statement += ' AND ';
        }
        statement += this._escapeName(this._partitionFieldNames(inTableName, [curCondition[0]])[0]) // We know it's not partitioned
            + curCondition[1] // Conditional operator is the same as in SimpleDb
            + this._escapeValue(curCondition[2]);
    }
    if (inOptions.orderBy) {
        statement += ' ORDER BY ' + this._escapeName(this._partitionFieldNames(inTableName, [inOptions.orderBy])[0]); // Already checked not partitioned
        if (inOptions.orderDir) {
            statement += ' ' + ((inOptions.orderDir.toLowerCase() === 'desc') ? 'desc' : 'asc');
        }
    }
    if (inOptions.limit) { // Purposely exclude if 0
        statement += ' LIMIT ' + parseInt(inOptions.limit);
    }
        
    // Run the query, wrapper uses NextToken to get all results
    var me = this;
    this._wrapSelect(
        statement,
        function (inError, inResult) {
        
            if (inError) {
                inCallback({code: 'SIMPLEDB_SELECT_ERROR', chain: inError});
            } else {
                            
                // Join the field parts back together and return
                var ret = [],
                    curId,
                    curItem;
                for (var i = 0; i < inResult.length; i++) {
                    curItem = inResult[i];
                    curId = curItem.$ItemName;
                    delete curItem.$ItemName;
                    curItem = me._unpartitionRow(inTableName, curItem);
                    curItem.__id = curId;
                    ret.push(curItem);
                }
                
                inCallback(null, ret);
                
            }
            
        }
    );
    
};


StringTabler.prototype.countQuery = function (inTableName, inCondition, inCallback) {

    // Validate condition (does not allow __id, if you had it you would getMulti)
    err = this._validateCondition(inTableName, inCondition);
    if (err) {
        inCallback({code: err});
        return;
    }
    
    // Build the select statement
    var statement = 'SELECT count(*) FROM ' + this._escapeName(this._getDomainName(inTableName)),
        curCondition;
    for (i = 0; i < inCondition.length; i++) {
        curCondition = inCondition[i];
        if (i === 0) {
            statement += ' WHERE ';
        } else {
            statement += ' AND ';
        }
        statement += this._escapeName(this._partitionFieldNames(inTableName, [curCondition[0]])[0]) // We know it's not partitioned
            + curCondition[1] // Conditional operator is the same as in SimpleDb
            + this._escapeValue(curCondition[2]);
    }
        
    // Run the query, wrapper uses NextToken to get all results
    var me = this;
    this._wrapSelect(
        statement,
        function (inError, inResult) {
        
            if (inError) {
                inCallback({code: 'SIMPLEDB_SELECT_ERROR', chain: inError});
            } else {
                
                // If nextTokens are used then we will have multiple
                // results -- iterate
                var count = 0,
                    i;
                for (i = 0; i < inResult.length; i++) {
                    count += parseInt(inResult[i].Count); // SDB always returns strings
                }
                inCallback(null, count);
                
            }
            
        }
    );
    
};


StringTabler.prototype.updateOne = function (inTableName,inRowWithoutId, inRowId, inCallback) {

    // Convenience function, special case of updateMulti
    this.updateMulti(inTableName, inRowWithoutId, [inRowId], function (inError) {
    
        if (inError) {
            inCallback(inError);
        } else {
            inCallback(null);
        }
        
    });

};

StringTabler.prototype.updateMulti = function (inTableName, inRowWithoutId, inRowIdArray, inCallback) {

    // Validate the table and fields
    var err = this._validateRowWithoutId(inTableName, inRowWithoutId, true); // Allow missing fields
    if (err) { 
        inCallback({code: err});
        return;
    }
    
    // Translate the write (i.e. partition)
    var tranRowWithoutId = this._partitionRow(inTableName, inRowWithoutId);
    
    // BatchPutAttributes creates items that don't exist, and using PutAttributes you
    // can only condition the write on the existence of an attribute, not the existence of the item
    // So start by using selectMulti to determine which items actually exist, and write only
    // those -- you lose speed, but this is needed for integrity
    var me = this;
    this.selectMulti(
        inTableName,
        [this._getShortestFieldName(inTableName)],
        inRowIdArray,
        function (inError, inRowWithIdHash) {
                
            if (inError) {
                inCallback({code: 'SELECTMULTI_ERROR', chain: inError});
            } else {
            
                // Create the argument for the SDB call, updating only
                // the items that actually exist. The SDB module doesn't support
                // replacing, so everything is done using the override.
                var rowId,
                    row,
                    override = {},
                    itemNum = 0,
                    attNum = 0,
                    attName;
                for (rowId in inRowWithIdHash) { // Only add the ones that exist (they will be in the hash)
                
                    override['Item.' + itemNum + '.ItemName'] = rowId;
                    attNum = 0;
                    for (attName in tranRowWithoutId) {
                        override['Item.' + itemNum + '.Attribute.' + attNum + '.Name']      = attName;
                        override['Item.' + itemNum + '.Attribute.' + attNum + '.Value']     = tranRowWithoutId[attName];
                        override['Item.' + itemNum + '.Attribute.' + attNum + '.Replace']   = "true"
                        attNum++;
                    }
                    itemNum++;
                }
                                
                // Run the API request
                me._wrapBatchPutItem( // Does parallel requests if there are >25 items
                    me._getDomainName(inTableName),
                    [], // Specified in override
                    override,
                    function (inError) {
                    
                        if (inError) {
                            inCallback({code: 'SIMPLEDB_BATCHPUTITEMS_ERROR', chain: inError});
                        } else {
                            inCallback(null);
                        }
                    
                    }
                );
                
            
            }
        
        
        }
    );
    
};

StringTabler.prototype.updateQuery = function (inTableName, inRowWithoutId, inCondition, inCallback) {

    // Convenience function (with some efficiency gain because you don't need to check
    // for row existence in the query results like updateSingle/Multi do).
    // Get the id of rows that match the query and then run the updates on those rows

    // Validate the table and fields
    var err = this._validateRowWithoutId(inTableName, inRowWithoutId, true); // Allow missing fields
    if (err) { 
        inCallback({code: err});
        return;
    }
        
    // Partition the write (only ince)
    var tranRowWithoutId = this._partitionRow(inTableName, inRowWithoutId);
        
    // The query is validated by selectQuery
    var me = this;
    this.selectQuery(
        inTableName,
        [],
        inCondition, // Validated by selectQuery
        function (inError, inRowWithIdArray) {
                
            if (inError) {
                inCallback(inError); // Intentionally do not chain
            } else {
            
                // Create the argument for the SDB call. The SDB module doesn't support
                // replacing, so everything is done using the override.
                var i,
                    curRow,
                    row,
                    override = {},
                    itemNum = 0,
                    attNum = 0,
                    attName;
                for (i = 0; i < inRowWithIdArray.length; i++) { // Only add the ones that exist (they will be in the hash)
                    curRow = inRowWithIdArray[i];
                    override['Item.' + itemNum + '.ItemName'] = curRow.__id;
                    attNum = 0;
                    for (attName in tranRowWithoutId) {
                        override['Item.' + itemNum + '.Attribute.' + attNum + '.Name']      = attName;
                        override['Item.' + itemNum + '.Attribute.' + attNum + '.Value']     = tranRowWithoutId[attName];
                        override['Item.' + itemNum + '.Attribute.' + attNum + '.Replace']   = "true"
                        attNum++;
                    }
                    itemNum++;
                }
                
                // Run the API request
                me._wrapBatchPutItem( // Does parallel requests if there are >25 items
                    me._getDomainName(inTableName),
                    [], // Specified in override
                    override,
                    function (inError) {
                    
                        if (inError) {
                            inCallback({code: 'SIMPLEDB_BATCHPUTITEMS_ERROR', chain: inError});
                        } else {
                            inCallback(null);
                        }
                    
                    }
                );
                
            
            }
        
        
        }
    );
    
};


StringTabler.prototype.deleteOne = function (inTableName, inRowId, inCallback) {

    // Convenience function, special case of deleteMulti
    this.deleteMulti(inTableName, [inRowId], function (inError) {
    
        if (inError) {
            inCallback(inError);
        } else {
            inCallback(null);
        }
        
    });

};

StringTabler.prototype.deleteMulti = function (inTableName, inRowIdArray, inCallback) {

    // Validate the table
    if (!this._tables[inTableName]) {
        inCallback({'code': 'NON_EXITENT_TABLE'});
        return;
    }
    
    // Create the argument
    var items = [],
        i = 0;
    for (i = 0; i < inRowIdArray.length; i++) {
        items.push({'$ItemName': inRowIdArray[i]});
    }
    
    // Run BatchDeleteAttributes and don't worry about whether or not the
    // rows ever actually existed
    // Use the wrapper to easily paralellize when the 25-item limit is reached
    // We are deleting entire items, so we don't need to specify the attributes
    this._wrapBatchDeleteItem( 
        this._getDomainName(inTableName),
        items,
        function (inError) {
        
            if (inError) {
                inCallback({code: 'SIMPLEDB_BATCHDELETEITEMS_ERROR', chain: inError});
            } else {
                inCallback(null);
            }
        
        }
    );
    
};

StringTabler.prototype.deleteQuery = function (inTableName, inCondition, inCallback) {

    // Convenience function
    // Get the id of rows that match the query and then delete those rows

    // Validate the table
    if (!this._tables[inTableName]) {
        inCallback({'code': 'NON_EXITENT_TABLE'});
        return;
    }
    
    // The query is validated by selectQuery
    var me = this;
    this.selectQuery(
        inTableName,
        [],
        inCondition, // Validated by selectQuery
        function (inError, inRowWithIdArray) {
                        
            if (inError) {
                inCallback(inError); // Intentionally do not chain
            } else {
            
                // Get array of row ids
                var rowIds = [],
                    i;
                for (i = 0; i < inRowWithIdArray.length; i++) {
                    rowIds.push(inRowWithIdArray[i].__id);
                }
            
                // Use deleteMulti to run the API request
                me.deleteMulti(
                    inTableName,
                    rowIds,
                    function (inError) {
                    
                        if (inError) {
                            inCallback(inError); // Purposely not chained
                        } else {
                            inCallback(null);
                        }
                    
                    }
                );
                
            
            }
        
        
        }
    );
    
};


// Private

StringTabler.prototype._getDomainName = function (inTableName) {

    // SDB component should worry about escaping this, and I do
    // it manually for selects
    return this._domainNamespace + '-' + inTableName;

};

StringTabler.prototype._escapeName = function (inValue) {

    // Table and attribute names

    return '`' + inValue.replace(/`/g, '``') + '`';
    
};

StringTabler.prototype._escapeValue = function (inValue) {

    // String values

    return "'" + inValue.replace(/'/g, "''") + "'";
    
};

StringTabler.prototype._partitionRow = function (inTableName, inRow) {

    /*
    
    Translate from client-friendly ({AAA: 'ab', b: 'b'}) to SDB-friendly
    ({AAA_1: 'a', a_2: 'b', ...}). Keep case used by client code, I prevented
    duplicates in defineTable().
    
    Assumes field names and values have already been validated.
    
    You need to return all parts for each field present (i.e. if a field is short you
    still need to write all unused parts to empty strings because we
    might be overwriting them here.
    
    But if a field isn't present at all in inRow we do NOT add it, this would
    break partial row updates.
        
    */
    
    var ret = {},
        k,
        i,
        parts;
        
    for (k in inRow) {
        parts = this._partitionString(inRow[k], 1024); // Split by bytes into UTF-8 valid pieces of at most 1024 bytes
        for (i = 0; i < this._tables[inTableName][k].numAttributes; i++) {
            ret[k + '__' + i] = parts[i] || ''
        }
    }
    
    return ret;

};

StringTabler.prototype._partitionFieldNames = function (inTableName, inFieldNames) {

    /*
    
    Translate from client-friendly ['AAA', 'b'] to SDB-friendly
    ['AAA_0' ,'AAA_1', 'b_0', ...]. Keep case used by client code, I prevented
    duplicates in defineTable().
    Assume field names and values have already been validated.
        
    */
    
    var ret = [],
        i,
        j, 
        curName,
        parts;
        
    for (i = 0; i < inFieldNames.length; i++) {
        curName = inFieldNames[i];
        for (j = 0; j < this._tables[inTableName][curName].numAttributes; j++) {
            ret.push(curName + '__' + j);
        }
    }
    
    return ret;

};

StringTabler.prototype._unpartitionRow = function (inTableName, inTranslatedRow) {

    // Inverse of _partitionRow()
    
    var ret = {},
        k,
        i,
        fieldName,
        splitValue;
        
    // Recomibine fields
    for (k in inTranslatedRow) {
        splitValue = k.split('__');
        fieldName = splitValue[0];
        
        if (!ret[fieldName]) {
            
            i = 0;
            ret[fieldName] = '';
            while (inTranslatedRow[fieldName + '__' + i]) {
                ret[fieldName] += inTranslatedRow[fieldName + '__' + i];
                i++;
            }
            
        }
    
    }
    
    return ret;
    
};

StringTabler.prototype._getShortestFieldName = function (inTableName) {

    // Return the shortest field in the table
    var smMaxByteLength = Infinity,
        smFieldName,
        fieldName,
        curField;
    for (fieldName in this._tables[inTableName]) {
        curField = this._tables[inTableName][fieldName];
        if (curField.maxByteLength < smMaxByteLength) {
            smMaxByteLength = curField.maxByteLength;
            smFieldName = fieldName;
        }
    }
    
    return smFieldName;
    
};

// Validators: return string error code or '' for success
// Assume the object structure is correct (it probably would be costly
// to verify every single object again the schema)

StringTabler.prototype._validateRowWithoutId = function (inTableName, inRowWithoutId, inAllowMissingFields) {

    var table,
        fieldName;
    
    // Verify the table
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Verify that the fields and values are consistent with the table definition.
    for (fieldName in table) { // fieldName key will be lower-case - good
        if (inRowWithoutId[fieldName] === undefined) {
            if (!inAllowMissingFields) {
                return 'MISSING_FIELD';
            }
        } else if (typeof inRowWithoutId[fieldName] !== 'string') {
            return 'VALUE_NOT_STRING';
        } else if (Util.stringByteLength(inRowWithoutId[fieldName]) > table[fieldName].maxByteLength) {
            return 'VALUE_TOO_LONG';
        }
    }
    for (fieldName in inRowWithoutId) {
        if (!this._tables[inTableName][fieldName]) {
            return 'EXTRANEOUS_FIELD';
        }
    }
    
    return '';

};

StringTabler.prototype._validateRowWithId = function (inTableName, inRowWithId, inAllowMissingFields) {

    var table,
        fieldName;
    
    // Verify the table
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Verify that the fields and values are consistent with the table definition.
    for (fieldName in table) { // fieldName key will be lower-case - good
        if (inRowWithId[fieldName] === undefined) {
            if (!inAllowMissingFields) {
                return 'MISSING_FIELD';
            }
        } else if (typeof inRowWithId[fieldName] !== 'string') {
            return 'VALUE_NOT_STRING';
        } else if (Util.stringByteLength(inRowWithId[fieldName]) > table[fieldName].maxByteLength) {
            return 'VALUE_TOO_LONG';
        }
    }
    for (fieldName in inRowWithId) {
        if (!this._tables[inTableName][fieldName] && fieldName !== '__id') {
            return 'EXTRANEOUS_FIELD';
        }
    }
    
    return '';

};

StringTabler.prototype._validateRowsWithoutIdHash = function (inTableName, inRowsWithoutIdHash, inAllowMissingFields) {

    var err,
        k;
    for (k in inRowsWithoutIdHash) {
        err = this._validateRowWithoutId(
            inTableName,
            inRowsWithoutIdHash[k],
            inAllowMissingFields
        );
        if (err) {
            return err;
        }
    }
    
    return '';
    
};

StringTabler.prototype._validateFieldNames = function (inTableName, inFieldNames) {

    // Field names can be empty (no fields) or null (all fields)
    // Make sure that the table exists, and that it has all these field names
    // Missing field names are always allowable
    // inAllowId indicates where the '__id' field is allowed
    
    var i,
        table,
        fieldName;
        
    // Verify the type
    if (!(inFieldNames instanceof Array) && inFieldNames !== null) {
        return 'INVALID_FIELD_LIST';
    }
    
    // Verify the table
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Verify that the fields and values are consistent with the table definition
    if (inFieldNames instanceof Array) {
        for (i = 0; i < inFieldNames.length; i++) {
            fieldName = inFieldNames[i];
            if (!this._tables[inTableName][fieldName]) {
                return 'NON_EXISTENT_FIELD';
            }
        }
    }
    
    return '';

};

StringTabler.prototype._validateCondition = function (inTableName, inCondition) {

    var table,
        condition,
        condFieldName,
        condOperator,
        condValue,
        i;
        
    if (!(table = this._tables[inTableName])) {
        return 'NON_EXISTENT_TABLE';
    }
    
    // Check the structure
    // Want to catch sketchily-defined conditions that result
    // in unintended writes
    if (!(inCondition instanceof Array)) {
        return 'INVALID_CONDITION';
    } else {
        for (i = 0; i < inCondition.length; i++) {
            if (!(inCondition[i] instanceof Array) || inCondition[i].length != 3) {
                return 'INVALID_CONDITION';
            }
        }
    }    
    
    // Check each condition
    for (i = 0; i < inCondition.length; i++) {
    
        condFieldName   = inCondition[i][0];
        condOperator    = inCondition[i][1];
        condValue       = inCondition[i][2];
        
        if (!table[condFieldName]) {
            return 'NON_EXISTENT_FIELD';
        } else if (table[condFieldName].numAttributes > 1) {
            return 'CANNOT_CONDITION_ON_PARTITIONED_FIELD';
        } else if (typeof condValue !== 'string') {
            return 'VALUE_NOT_STRING';
        } else if (Util.stringByteLength(condValue) > table[condFieldName].maxByteLength) {
            return 'VALUE_TOO_LONG';
        } else if (0 > ['=', '!=', '<', '>', '<=', '>='].indexOf(condOperator)) {
            return 'INVALID_OPERATOR';
        }
    
    }    
    
    return '';

};


// SimpleDB wrappers

StringTabler.prototype._wrapBatchPutItem = function () {

    // SDB can only put up to 25 items at once, so do parallel batches

    // Args (domain, items, cb) or (domain, items, overwrite, cb)
    var inDomain = arguments[0],
        inItems = arguments[1],
        inOverride = {},
        inCallback;
    if (arguments.length === 3) {
        inCallback = arguments[2];
    } else {
        inOverride = arguments[2];
        inCallback = arguments[3];
    }
        
    // Update requests will have the items in the override argument,
    // since that is the only way to replace attributes
    // So first things first, get all updates in the override argument
    var i,
        nextItemNum = 0,
        nextAttNum,
        k,
        parts;
    while (inOverride['Item.' + nextItemNum + '.ItemName']) {
        nextItemNum++;
    }
    for (i = 0; i < inItems.length; i++) {
        inOverride['Item.' + nextItemNum + '.ItemName'] = inItems[i].$ItemName;
        nextAttNum = 0;
        for (k in inItems[i]) {
            if (k !== '$ItemName') {
                inOverride['Item.' + nextItemNum + '.Attribute.' + nextAttNum + '.Name'] = k;
                inOverride['Item.' + nextItemNum + '.Attribute.' + nextAttNum + '.Value'] = inItems[i][k];
                // None of these are replace
                nextAttNum++;
            }
        }
        nextItemNum++;
    }
    
    // Now all the writes are in inOverride, so do batches using that
    // Create the 25-item slices of override arguments
    var overrideSlices = [],
        curSliceNum,
        curItemNum = 0,
        newItemNum,
        curAttNum;
    while (inOverride['Item.' + curItemNum + '.ItemName']) {
    
        curSliceNum = Math.floor(curItemNum / 25);
        newItemNum = curItemNum % 25;
        overrideSlices[curSliceNum] = overrideSlices[curSliceNum] || {};
        
        overrideSlices[curSliceNum]['Item.' + newItemNum + '.ItemName'] = inOverride['Item.' + curItemNum + '.ItemName'];
        
        curAttNum = 0;
        while (inOverride['Item.' + curItemNum + '.Attribute.' + curAttNum + '.Name']) {
            overrideSlices[curSliceNum]['Item.' + newItemNum + '.Attribute.' + curAttNum + '.Name']
                = inOverride['Item.' + curItemNum + '.Attribute.' + curAttNum + '.Name'];
            overrideSlices[curSliceNum]['Item.' + newItemNum + '.Attribute.' + curAttNum + '.Value']
                = inOverride['Item.' + curItemNum + '.Attribute.' + curAttNum + '.Value'];
            if (inOverride['Item.' + curItemNum + '.Attribute.' + curAttNum + '.Replace']) {
                overrideSlices[curSliceNum]['Item.' + newItemNum + '.Attribute.' + curAttNum + '.Replace'] = "true";
            }
            curAttNum++;
        }
        
        curItemNum++;
        
    }
    
    // Run the parallel requests
    var me = this;
    Async.forEach( // Parallel
        overrideSlices,
        function (inOverride, cb) {
        
            me._sdb.batchPutItem(inDomain, [], inOverride, function(inError) {
       
                if (inError) {
                    cb(inError);
                } else {
                    cb();
                }
                
            });
            
        }, function (inError) {
            
            if (inError) {
                inCallback({code: 'SIMPLEDB_BATCHPUTITEM_ERROR', chain: inError});
            } else {
                inCallback();
            }
        
        }
    );
    
  
  
};

StringTabler.prototype._wrapBatchDeleteItem = function (inDomain, inItems, inCallback) {

    // SDB can only delete up to 25 items at once, so do parallel batches
    // Doesn't support overriding because I don't need it and it makes things complex
    
    // Create the item slices
    var itemSlices = [],
        i;
        
    for (i = 0; i < inItems.length; i+= 25) {
        itemSlices.push(inItems.slice(i, i + 25));
    }
    
    // Run the parallel requests
    var me = this;
    Async.forEach( // Parallel
        itemSlices,
        function (inItemSlice, cb) {
            
            me._sdb.batchDeleteItem(inDomain, inItemSlice, function(inError) {
       
                if (inError) {
                    cb(inError);
                } else {
                    cb();
                }
                
            });
            
        }, function (inError) {
            
            if (inError) {
                inCallback({code: 'SIMPLEDB_BATCHDELETEITEM_ERROR', chain: inError});
            } else {
                inCallback();
            }
        
        }
    );
    
  
  
};

StringTabler.prototype._wrapSelect = function (inStatement, inCallback) {

    // Long queries might return a nextToken, which is
    // automatically iterated on here
    
    // Always use a LIMIT, since otherwise it defaults to 100
    
    // Note that LIMITs do not work will with next tokens, since
    // if a LIMIT restricts the result set then a nextToken
    // will always be issued, and when you retrieve data based on
    // nextToken its limit doesn't take into account previously
    // retrieved data. So you need to manually set the limit
    // each time based on the number of results that you STILL NEED
    var numStillNeeded = null; // No limit, here 0 means no more needed, not get all
    var regex = / limit *[0-9]+ *$/i;
    if (regex.test(inStatement)) {
        var parts = inStatement.split(/limit/gi);
        numStillNeeded = parseInt(parts[parts.length - 1]);
        inStatement = inStatement.replace(regex, '');
    }
    
    var done = false,
        nextToken,
        override = {},
        results = [],
        me = this;
    Async.whilst(
        function () { return !done; },
        function (cb) {
        
            me._sdb.select(
                inStatement + ' LIMIT ' + ((numStillNeeded !== null) ? Math.min(numStillNeeded, 2500) : 2500),
                override,
                function(inError, inResults, inMeta) {
                                
                    if (inError) {
                        cb(inError);
                    } else {
                    
                        // Append the results
                        results = results.concat(inResults);
                        if (typeof numStillNeeded === 'number') {
                            numStillNeeded -= inResults.length;
                        }
                        
                        // Are we done? Run into either user limit or the
                        // end of the data
                        nextToken = inMeta.result.SelectResult.NextToken;
                        if ((numStillNeeded === null || numStillNeeded > 0) && nextToken) {
                            override.NextToken = nextToken.replace(/\n/g, '');
                            done = false;
                        } else {
                            done = true;
                        }
                        
                        cb();
                    }
                    
                }
            );
        
        },
        function (inError) {
        
            if (inError) {
                inCallback({code: 'SIMPLEDB_SELECT_FAILED', chain: inError});
            } else {
                inCallback(null, results);
            }
            
        }
    );
    
  
  
};




// String functions don't really belong in this class, but oh well

StringTabler.prototype._partitionString = function (inString, inMaxBytesPerPart) {

    // This is used to break UTF-8 strings into partitions  with the
    // byte length of each part not exceeding inMaxBytesPerPart
    // This is because simpledb attributes have a maximum byte length, which
    // is different than characters for UTF-8
    
    var stringLeft = inString,
        buffer = new Buffer(inMaxBytesPerPart),
        bytesWritten,
        chunk,
        chunks = [];
        
    while (stringLeft.length > 0) {
    
        bytesWritten = buffer.write(stringLeft, 'utf8'); // Will not write partial characters if too long (see doc)
        chunks.push(chunk = buffer.toString('utf8', 0, bytesWritten));
        stringLeft = stringLeft.substr(chunk.length);
        
    }
    
    return chunks;
    
};








module.exports = StringTabler;
