var StringTabler_SimpleDb = require('./stringtabler_simpledb'),
    StringTabler_JsonFile = require('./stringtabler_jsonfile'),
    JsonSchemaValidator = require('json-schema'),
    Util = require('./util');
    
    
    
    

    
    
    
/*

Extends StringTabler by providing data types and null values. 
Values are "serialized" (and as part of that, validated) before being 
used with the StringTable. So each data type has a method of
verifying/serializing and unserializing. The interaction between this
object and client code takes place using numbers, dates, objects, whatever,
this object knows how to stringify it.

Where possible, error checking is done at the StringTable level for things
other than data types.

*/


var Tabler = function (inBackend, inBackendParams) {

    if (inBackend === 'jsonfile') {
        this._stringTabler = new StringTabler_JsonFile(inBackendParams);
    } else if (inBackend = 'simpledb') {
        this._stringTabler = new StringTabler_SimpleDb(inBackendParams);
    }
    
    /*
    
    This holds the field definitions in the form
    this._tables[tableName] = inFieldDefs
    
    */
    this._tables = {};
  
};

/*

Using the null byte to represent null is perfect (meaningful,
fits in single-byte fields, orders as you would want) except
that it triggers base64 encoding of return values. But only
the null value itself is encoded this way, and since it is
tiny the cost in bandwidth and processing is basically nothing.

*/

Tabler.prototype.NULL_STRING = "\u0000";

Tabler.prototype.defineTable = function (inTableName, inFieldDefs, inCallback) {

    // Translate inFieldDefs to the format used by StringTable
    var k,
        curDef,
        curDataType,
        stringDefs = {};
    for (k in inFieldDefs) {
        curDef = inFieldDefs[k];
        curDataType = this._dataTypes[curDef.type];
        if (curDataType && curDataType.isValidFieldDef(curDef)) {
            stringDefs[k] = curDataType.getDefByteLength(curDef);
        } else {
            inCallback({code: 'INVALID_FIELD_DEFINITION', fieldName: k});
            return;
        }
    }
    
    // Create the StringTable
    var me = this;
    this._stringTabler.defineTable(
        inTableName,
        stringDefs,
        function (inError) {
        
            if (inError) {
                inCallback(inError);
            } else {
          
                // If no error then save the field definitions
                me._tables[inTableName] = inFieldDefs;
                inCallback(null);
                
            }
        }
    );

};

Tabler.prototype.deleteTable = function (inTableName, inCallback) {

    // Delete from the StringTable
    var me = this;
    this._stringTabler.deleteTable(
        inTableName,
        function (inError) {
            
            if (inError) {
                inCallback(inError);
            } else {
                delete me._tables[inTableName];
                inCallback(null);
            }
            
        }
    );
    
};


Tabler.prototype.insertOne = function (inTableName, inRowWithoutId, inCallback) {

    // Try to serialize the field values
    var serRow;
    try {
        serRow = this._serializeRow(inTableName, inRowWithoutId);
    } catch (e) {
        inCallback(e);
        return;
    }
    
    // Try to insert to string table
    this._stringTabler.insertOne(
        inTableName,
        serRow,
        inCallback
    );
    
};

Tabler.prototype.insertMulti = function (inTableName, inRowWithoutIdArray, inCallback) {

    // Try to serialize the field values
    var serRowArray = [],
        i,
        serRow;
    for (i = 0; i < inRowWithoutIdArray.length; i++) {
        try {
            serRow = this._serializeRow(inTableName, inRowWithoutIdArray[i]);
        } catch (e) {
            inCallback(e);
            return;
        }
        serRowArray.push(serRow);
    }
    
    // Try to insert to string table
    this._stringTabler.insertMulti(
        inTableName,
        serRowArray,
        inCallback
    );
    
};


Tabler.prototype.selectOne = function (inTableName, inFieldNames, inRowId, inCallback) {

    // Run the select on the string table
    var me = this;
    this._stringTabler.selectOne(
        inTableName,
        inFieldNames,
        inRowId,
        function (inError, inRowWithoutId) {
        
            if (inError) {
                inCallback(inError);
            } else {

                if (inRowWithoutId === null) {
                    inCallback(null, null); // Not found, don't unserialize
                } else {
            
                    // Unserialize the values
                    inCallback(
                        null,
                        me._unserializeRow(inTableName, inRowWithoutId)
                    );

                }
                
            }
        
        
        }
    );

};

Tabler.prototype.selectMulti = function (inTableName, inFieldNames, inRowIdArray, inCallback) {

    // Run the select on the string table
    var me = this;
    this._stringTabler.selectMulti(
        inTableName,
        inFieldNames,
        inRowIdArray,
        function (inError, inRowWithIdHash) {
        
            if (inError) {
                inCallback(inError);
            } else {
            
                // Unserialize
                var k,
                    unserHash = {};
                for (k in inRowWithIdHash) {
                    unserHash[k] = me._unserializeRow(inTableName, inRowWithIdHash[k]);
                }
                inCallback(null, unserHash);
                
            }
        
        
        }
    );

};

Tabler.prototype.selectQuery = function (inTableName, inFieldNames, inCondition) {

    // Arguments
    var inOrderBy,
        inCallback;
    if (arguments.length === 4) {
        inOptions = {};
        inCallback = arguments[3];
    } else if (arguments.length === 5) {
        inOptions = arguments[3];
        inCallback = arguments[4];
    }

    // Try to serialize the condition values
    var serCond;
    try {
        serCond = this._serializeCondition(inTableName, inCondition);
    } catch (e) {
        inCallback(e);
        return;
    }
        
    // Run the select on the string table
    var me = this;
    this._stringTabler.selectQuery(
        inTableName,
        inFieldNames,
        serCond,
        inOptions,
        function (inError, inRowsWithIdArray) {
        
            if (inError) {
                inCallback(inError);
            } else {
            
                // Unserialize
                var i,
                    unserArr = [];
                for (i = 0; i < inRowsWithIdArray.length; i++) {
                    unserArr.push(
                        me._unserializeRow(inTableName, inRowsWithIdArray[i])
                    );
                }
                inCallback(null, unserArr);
                
            }
        
        
        }
    );

};


Tabler.prototype.countQuery = function (inTableName, inCondition, inCallback) {


    // Try to serialize the condition values
    var serCond;
    try {
        serCond = this._serializeCondition(inTableName, inCondition);
    } catch (e) {
        inCallback(e);
        return;
    }
    
    // Run the select on the string table
    var me = this;
    this._stringTabler.countQuery(
        inTableName,
        serCond,
        inCallback // Don't need to unserialize the count
    );
};


Tabler.prototype.updateOne = function (inTableName, inRowWithoutId, inRowId, inCallback) {

    // Try to serialize the field values
    var serRow;
    try {
        serRow = this._serializeRow(inTableName, inRowWithoutId);
    } catch (e) {
        inCallback(e);
        return;
    }
    
    // Run the operation
    this._stringTabler.updateOne(
        inTableName,
        serRow,
        inRowId,
        inCallback
    );
    
};

Tabler.prototype.updateMulti = function (inTableName, inRowWithoutId, inRowIdArray, inCallback) {

    // Try to serialize the field values
    var serRow;
    try {
        serRow = this._serializeRow(inTableName, inRowWithoutId);
    } catch (e) {
        inCallback(e);
        return;
    }
    
    // Run the operation
    this._stringTabler.updateMulti(
        inTableName,
        serRow,
        inRowIdArray,
        inCallback
    );
    
};

Tabler.prototype.updateQuery = function (inTableName, inRowWithoutId, inCondition, inCallback) {

    // Try to serialize the field values
    var serRow;
    try {
        serRow = this._serializeRow(inTableName, inRowWithoutId);
    } catch (e) {
        inCallback(e);
        return;
    }
    
    // Try to serialize the condition values
    var serCond;
    try {
        serCond = this._serializeCondition(inTableName, inCondition);
    } catch (e) {
        inCallback(e);
        return;
    }
    
    // Run the operation
    this._stringTabler.updateQuery(
        inTableName,
        serRow,
        serCond,
        inCallback
    );
    
};


Tabler.prototype.deleteOne = function (inTableName, inRowId, inCallback) {

    // Run the operation
    this._stringTabler.deleteOne(
        inTableName,
        inRowId,
        inCallback
    );
    
};

Tabler.prototype.deleteMulti = function (inTableName, inRowIdArray, inCallback) {

    // Run the operation
    this._stringTabler.deleteMulti(
        inTableName,
        inRowIdArray,
        inCallback
    );
    
};

Tabler.prototype.deleteQuery = function (inTableName, inCondition, inCallback) {
    
    // Try to serialize the condition values
    var serCond;
    try {
        serCond = this._serializeCondition(inTableName, inCondition);
    } catch (e) {
        inCallback(e);
        return;
    }

    // Run the operation
    this._stringTabler.deleteQuery(
        inTableName,
        serCond,
        inCallback
    );
    
};



// String patterns for convenience (they are required, remember -- make it easy to reuse them)

Tabler.prototype.PATTERNS = {

    // Value must match pattern to be valid

    // Allow absolutely anything
    ALL: /.*/,
    
    // Disallow C0 and C1 control codes and DEL (127).
    // http://en.wikipedia.org/wiki/C0_and_C1_control_codes
    PRINTABLE_NO_BREAKS: /^[^\u0000-\u001f\u007f-\u009f]*$/,
    
    // Same but allow line break characters
    PRINTABLE_AND_BREAKS: /^[^\u0000-\u0009\u000B\u000C\u000E-\u001f\u007f-\u009f]*$/
    
};



// Private

Tabler.prototype._serializeRow = function (inTableName, inRow) {

    // Do not try to serialize __ids if they are there, but keep them

    if (!this._tables[inTableName]) {
        throw {code: 'NON_EXISTENT_TABLE'};
    }
    
    var serRow = {},
        k,
        curDef;
    for (k in inRow) {
        if (k === '__id') { 
            serRow[k] = inRow[k];
        } else {
            curDef = this._tables[inTableName][k];
            if (!curDef) {
                throw {code: 'INVALID_FIELD_NAME', fieldName: k};
            }
            try {
                serRow[k] = this._serializeField(curDef, inRow[k]);
            } catch (e) {
                e.fieldName = k;
                throw e;
            }
        }
    }

    return serRow;
    
};

Tabler.prototype._unserializeRow = function (inTableName, inRow) {

    // Ignore __id rows, we leave them as strings

    if (!this._tables[inTableName]) {
        throw {code: 'NON_EXISTENT_TABLE'};
    }
    
    var unserRow = {},
        k,
        curDef,
        curType,
        curVal;
    for (k in inRow) {
        if (k === '__id') {
            unserRow['__id'] = inRow[k];
        } else {
            curDef = this._tables[inTableName][k];
            if (!curDef) {
                throw {code: 'INVALID_FIELD_NAME', fieldName: k};
            }
            
            if (inRow[k] === this.NULL_STRING) {
                curVal = null;
            } else {
                curType = this._dataTypes[curDef.type];
                try {
                    curVal = curType.unserialize(curDef, inRow[k]);
                } catch (e) {
                    e.fieldName = k;
                    throw e;
                }
            }
            unserRow[k] = curVal;
        }
    }

    return unserRow;
    
};

Tabler.prototype._serializeCondition = function (inTableName, inCondition) {

    if (!this._tables[inTableName]) {
        throw {code: 'NON_EXISTENT_TABLE'};
    }

    var serCond = [],
        serVal,
        i,
        curFieldDef,
        curType;
    for (i = 0; i < inCondition.length; i++) {
        curFieldDef = this._tables[inTableName][inCondition[i][0]];
        if (!curFieldDef) {
            throw {code: 'INVALID_FIELD_NAME', fieldName: inCondition[i][0]};
        }
        curType = this._dataTypes[curFieldDef.type];
        try {
            serVal = this._serializeField(curFieldDef, inCondition[i][2]);
        } catch (e) {
            e.fieldName = inCondition[i][0];
            throw e;
        }
        serCond.push([
            inCondition[i][0],
            inCondition[i][1],
            serVal
        ])
    }
    
    return serCond;

};

Tabler.prototype._serializeField = function (inFieldDef, inValue) {
    
    // Handle nulls here, or if not null then pass off to 
    // type-specific serialization

    var curString,
        curType;
    if (inValue === null || inValue === this.NULL_STRING) { // Prevent "covert" null insertion
        if (inFieldDef.nullable) {
            curString = this.NULL_STRING;
        } else {
            throw {code: 'NOT_NULLABLE'};
        }
    } else {
        curType = this._dataTypes[inFieldDef.type];
        curString = curType.serialize(inFieldDef, inValue); // Throws exception if no good -- don't catch
    }
    return curString;

};

// Data types

/*

Each data type is an object with the following functions:

isFieldDefValid             (fieldDef)
Return true/false according to whether required parameters are there
and valid.

getSerializedByteLength     (fieldDef)
Get the maximum byte length that will be required by a serialized value.

serialize                   (fieldDef, value)
Get the stringified value of the object, or return false
if not valid.

unserialize                 (fieldDef, string)
Restore the original object.

The fieldDef parameter has all the information needed to
validate

Deserialize functions can rely on the fact that the data
was serialized properly, but must also return null values
where appropriate.

*/

Tabler.prototype._dataTypes = {};

Tabler.prototype._dataTypes.string = {

    isValidFieldDef: function (inFieldDef) {
    
        return (
            typeof inFieldDef.byteLength === "number"
            && inFieldDef.byteLength === parseInt(inFieldDef.byteLength)
            && inFieldDef.byteLength > 0
            && inFieldDef.pattern instanceof RegExp
        );
    
    },

    getDefByteLength: function (inFieldDef) {
    
        return inFieldDef.byteLength;
    
    },
    
    serialize: function (inFieldDef, inValue) {
    
        // Validate
        if (
            typeof inValue !== 'string'
            || Util.stringByteLength(inValue) > inFieldDef.byteLength
            || !inFieldDef.pattern.test(inValue)
        ) {
            throw {code: 'INVALID_VALUE'};
        }
        
        // Stringify
        return inValue;
        
    },
    
    unserialize: function (inFieldDef, inString) {
    
        return inString;
    
    }

};

Tabler.prototype._dataTypes.number = {

    isValidFieldDef: function (inFieldDef) {
    
        return (
            typeof inFieldDef.decPlaces === "number" // Number of decimal places
            && inFieldDef.decPlaces === parseInt(inFieldDef.decPlaces)
            &&typeof inFieldDef.min === "number" // These do not have to be integers
            && typeof inFieldDef.max === "number"
        );
    
    },

    getDefByteLength: function (inFieldDef) {
    
        // The offset (added) is -min
        // The max number we will need to actually store is then max-min
        // And there can be a maximum of .decPlaces decimal places
        
        // This is what an integer would use
        var intBytes = Math.floor(Math.log(inFieldDef.max - inFieldDef.min) / Math.log(10)) + 1 + 1;
       
        // Add the decimal point and the decPlaces
        return intBytes + 1 + inFieldDef.decPlaces; 
        
    },
    
    serialize: function (inFieldDef, inValue) {
    
        // Have to use offsetting and zero-padding so that lexicographical
        // order is the same as numerical order
        
        // You don't need to pad the decimal places, but you do
        // need to use rounding to ensure that they are stored to the
        // correct precision (i.e. fit in the field)
                
        var parts = (inValue + '').split('.'),
            dec = parts[1] || '';
        if (
            typeof inValue !== 'number'
            || inValue > inFieldDef.max
            || inValue < inFieldDef.min
            || /[^\.\-0-9]/.test('' + inValue) // Testing for scientific notation
            || dec.length > inFieldDef.decPlaces // Make sure we can store this number
        ) {
            throw {code: 'INVALID_VALUE'};
        }
    
        // If the minimum value is negative then we need to offset by that amount
        var offset = -inFieldDef.min; // This can be negative or positive
                
        // The maximum number we could be storing is then max + offset so zero pad to support that
        var maxStored = inFieldDef.max + offset,
            charsNeededBeforeDecimal = Math.floor(Math.log(maxStored) / Math.log(10)) + 1 + 1; // Js only has base-2 log. There is an extra +1 because sometimes floating point math makes the floor argument .99999 instead of .0 - be safe
            storeValue = '' + (inValue + offset);
            decPos = (storeValue.indexOf('.') < 0) // Not needed here -- ints only
                ? storeValue.length
                : storeValue.indexOf('.'); // This many zeros are NOT needed
            zerosToAdd = charsNeededBeforeDecimal - decPos;
            ret = '' + (new Array(zerosToAdd + 1)).join('0') + storeValue;
        
        return ret;
        
    },
    
    unserialize: function (inFieldDef, inString) {
    
        // Kill the zero padding and add back the offset
        // Round to precision because adding back the offset seems to
        // introduce .00000000000x
        return Util.round((1 * inString) + inFieldDef.min, inFieldDef.decPlaces);
    
    }

};

Tabler.prototype._dataTypes.datetime = {

    isValidFieldDef: function (inFieldDef) {
    
        return true; // No params
    
    },

    getDefByteLength: function (inFieldDef) {
    
        return 20; // 2011-11-14T13:23:33Z
        
    },
    
    serialize: function (inFieldDef, inValue) {
    
        // inValue is a date object
        // Always serialize the date in UTC so that
        // comparisons are valid
    
        if (
            !(inValue instanceof Date)
        ) {
            throw {code: 'INVALID_VALUE'};
        }
    
        return inValue.getUTCFullYear()
            + '-' + Util.pad((inValue.getUTCMonth() + 1), 2, '0')
            + '-' + Util.pad(inValue.getUTCDate()       , 2, '0')
            + 'T' + Util.pad(inValue.getUTCHours()      , 2, '0')
            + ':' + Util.pad(inValue.getUTCMinutes()    , 2, '0')
            + ':' + Util.pad(inValue.getUTCSeconds()    , 2, '0')
            + 'Z';
        
    },
    
    unserialize: function (inFieldDef, inString) {
    
        var parts = inString.split('T'),
            dateParts = parts[0].split('-'),
            timeParts = parts[1].replace('Z', '').split(':');

        var myDate = new Date();
        myDate.setUTCFullYear(  dateParts[0]);
        myDate.setUTCMonth(     dateParts[1] - 1);
        myDate.setUTCDate(      dateParts[2]);
        myDate.setUTCHours(     timeParts[0]);
        myDate.setUTCMinutes(   timeParts[1]);
        myDate.setUTCSeconds(   timeParts[2]);
        myDate.setUTCMilliseconds(0);

        return myDate;

    }

};

Tabler.prototype._dataTypes.object = {

    // Very similar to string except that you pass objects (or anything
    // JSON-encodable

    isValidFieldDef: function (inFieldDef) {
    
        return (
            typeof inFieldDef.byteLength === "number"
            && inFieldDef.byteLength === parseInt(inFieldDef.byteLength)
            && inFieldDef.byteLength > 0
        );
        
    },

    getDefByteLength: function (inFieldDef) {
    
        return inFieldDef.byteLength;
        
    },
    
    serialize: function (inFieldDef, inValue) {
    
        // Get JSON string
        var json;
        try {
            json = JSON.stringify(inValue);
        } catch (e) {
            throw {code: 'INVALID_VALUE'};
        }
        if (typeof json !== 'string') { // For example, JSON.stringify(function () {}) === undefined
            throw {code: 'INVALID_VALUE'};
        }
        
        // Validate
        if (
            Util.stringByteLength(json) > inFieldDef.byteLength
            || (inFieldDef.schema && !JsonSchemaValidator.validate(inValue, inFieldDef.schema).valid)
        ) {
            throw {code: 'INVALID_VALUE'};
        }
        
        return json;
        
    },
    
    unserialize: function (inFieldDef, inString) {
    
        return JSON.parse(inString);
        
    }

};








module.exports = Tabler;
