﻿// You pass a command line argument of the file that the
// module is in
if (process.argv.length < 3) {
    console.log('Run something like: node stringtable.js jsonfile');
    process.exit();
}
if (process.argv[2] === 'simpledb' && process.argv.length !== 5) {
    console.log('Run something like: node stringtable.js simpledb keyid key');
    process.exit();
}

// Module
var vows = require('vows'),
    assert = require('assert'),
    StringTabler = require('../lib/stringtabler_' + process.argv[2]),
    Util = require('../lib/util'),
    fs = require('fs');
    
    
    

var suite = vows.describe('stringtable');


// This data should be valid
var myTabler,
    myTablerName = 'String_Table_Test',
    myFieldDefs = {
        'username': 20,
        'post':     1500
    },
    myFieldName1 = 'username', // Should be one valid field name for single-field gets
    myFieldName2 = 'post', // Should be the long field that causes partitioning
    myRow1 = { // Rows should be alphabetical on username
        'username': '\'"`Aaron', // Require escaping
        'post':     '\'"`W' + (new Array(1400)).join('a') // Require escaping
    },
    myRow2 = {
        'username': 'Gill',
        'post':     'W' + (new Array(1500)).join('b')
    },
    myRow3 = {
        'username': 'Joe',
        'post':     'W' + (new Array(1500)).join('c')
    },
    myRow1Id1,
    myRow1Id2,
    myRow1Id3,
    myBigNum = 3000; // Number of rows to insert for large data tests
    
suite.addBatch({
    
    'instantiate the object': {
    
        'topic': function () {
            if (process.argv[2].toLowerCase() === 'simpledb') {
                myTabler = new StringTabler({
                    accessKeyId:        process.argv[3],
                    secretAccessKey:    process.argv[4],
                    domainNamespace:    'test'
                });
            } else {
                myTabler = new StringTabler({
                    filename: 'stringtable.json'
                });
            }   
            return true;
        },
        
        'worked': function () {
            assert.isObject(myTabler);
        }
        
    }
        
})
.addBatch({
    
    'try to define an invalid table': {
    
        'topic': function () {
            myTabler.defineTable(
                myTablerName,
                {
                    'User`name':          'joke'
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_DEFINITION');
        }
  
    }
})
.addBatch({
    
    'try to define another invalid table': {
    
        'topic': function () {
            myTabler.defineTable(
                myTablerName,
                {
                    'User`name':         10,
                    'user`name':         10, // Don't allow multiple fields that with the same lowercase name
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'DUPLICATE_FIELD_NAMES');
        }
  
    }
})
.addBatch({
    
    'try to define a table with an unacceptably long row': {
    
        'topic': function () {
            myTabler.defineTable(
                myTablerName,
                {
                    'User`name':         260866, // 260865 = 1019 * 256
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'ROW_TOO_LONG');
        }
  
    }
})
.addBatch({
    
    'try to define a valid table': {
    
        'topic': function () {
            myTabler.defineTable(
                myTablerName,
                myFieldDefs,
                this.callback
            );
        },
        
        'and it works': function (inError) {
            assert.equal(inError, null); // Not strict
        }
  
    }
    
})
.addBatch({
        
    'try to define a duplicate table': {
    
        'topic': function () {
            myTabler.defineTable(
                myTablerName,
                myFieldDefs,
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'TABLE_ALREADY_DEFINED');
        }
        
    }

})
.addBatch({
        
    'try to insert a row into a non-existent table': {
    
        'topic': function () {
            myTabler.insertOne(
                myTablerName + '_junk',
                {
                    'User`name':          10000
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'NON_EXISTENT_TABLE');
        }
        
    }

})
.addBatch({
        
    'try to insert a row with an extraneous field': {
    
        'topic': function () {
            var row = Util.deepCopy(myRow1);
            row.garbage_field123 = 'test';
            myTabler.insertOne(
                myTablerName,
                row,
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'EXTRANEOUS_FIELD');
        }
        
    }

})
.addBatch({
        
    'try to insert a row with a missing field': {
    
        'topic': function () {
            myTabler.insertOne(
                myTablerName,
                {},
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'MISSING_FIELD');
        }
        
    }

})
.addBatch({
        
    'try to insert a row with a bad value type': {
    
        'topic': function () {
            var row = Util.deepCopy(myRow1),
                k;
            for (var k in row) {
                row[k] = 1;
            }
        
            myTabler.insertOne(
                myTablerName,
                row,
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'VALUE_NOT_STRING');
        }
        
    }

})
.addBatch({
        
    'try to insert a row with a too-long string': {
    
        'topic': function () {
            var row = Util.deepCopy(myRow1),
                k;
            for (k in row) {
                row[k] = (new Array(myFieldDefs[k] + 2)).join('a')
            }
            myTabler.insertOne(
                myTablerName,
                row,
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'VALUE_TOO_LONG');
        }
        
    }

})
.addBatch({
        
    'try to insert a valid row': {
    
        'topic': function () {
            myTabler.insertOne(
                myTablerName,
                myRow1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowId) {
            assert.equal    (inError, null); // Not strict
            assert.isString (inRowId);
            myRowId1 = inRowId;
        }
        
    }

})
.addBatch({
        
    'try to insert two valid rows': {
    
        'topic': function () {
            myTabler.insertMulti(
                myTablerName,
                [
                    myRow2,
                    myRow3
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowIdArray) {
            assert.equal    (inError, null); // Not strict
            assert.isArray  (inRowIdArray);
            assert.equal    (inRowIdArray.length, 2);
            assert.isString (inRowIdArray[0]);
            assert.isString (inRowIdArray[1]);
            myRowId2 = inRowIdArray[0];
            myRowId3 = inRowIdArray[1];
        }
        
    }

})
.addBatch({
        
    'try to get one non-existent row': {
    
        'topic': function () {
            myTabler.selectOne(
                myTablerName,
                [],
                'some_junk_row_id',
                this.callback
            );
        },
        
        'and the request succeeds but null is returned': function (inError, inRowWithoutId) {
            assert.isNull     (inError);
            assert.isNull     (inRowWithoutId);
            
        }
        
    }

})
.addBatch({
        
    'try to get one valid row, all fields': {
    
        'topic': function () {
            myTabler.selectOne(
                myTablerName,
                null,
                myRowId1,
                this.callback
            );
        },
        
        'and you get that row': function (inError, inRowWithoutId) {
        
            assert.isNull   (inError);
            assert.deepEqual (inRowWithoutId, myRow1);
            
        }
        
    }

})
.addBatch({
        
    'try to get one valid row with an invalid field': {
    
        'topic': function () {
            myTabler.selectOne(
                myTablerName,
                ['invalid_field_name'],
                myRowId1,
                this.callback
            );
        },
        
        'and you get an error': function (inError, inRowWithoutId) {
        
            assert.isObject     (inError);
            assert.equal        (inError.code, 'NON_EXISTENT_FIELD');
            
        }
        
    }

})
.addBatch({
        
    'try to get one valid row with a valid field specified': {
    
        'topic': function () {
            myTabler.selectOne(
                myTablerName,
                [myFieldName1],
                myRowId1,
                this.callback
            );
        },
        
        'and you get a row with that field only': function (inError, inRowWithoutId) {
            assert.isNull     (inError);
            for (var k in inRowWithoutId) {
                assert.equal (k, myFieldName1);
            }
        }
        
    }

})
.addBatch({
        
    'try to get one valid row with no fields (existence)': {
    
        'topic': function () {
            myTabler.selectOne(
                myTablerName,
                [], // No fields
                myRowId1,
                this.callback
            );
        },
        
        'and you get an empty object': function (inError, inRowWithoutId) {
        
            assert.isNull(inError);
            assert.isObject(inRowWithoutId);
            var k,
                count = 0;
            for (k in inRowWithoutId) {
                count++;
            }
            assert.strictEqual(count, 0);
            
        }
        
    }

})
.addBatch({
        
    'try to get two rows with an invalid field list': {
    
        'topic': function () {
            myTabler.selectMulti(
                myTablerName,
                ['some_junk_field'],
                [myRowId1, myRowId2],
                this.callback
            );
        },
        
        'and and it fails': function (inError, inRowWithIdHash) {
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'NON_EXISTENT_FIELD');
        }
        
    }

})
.addBatch({
        
    'try to get two existing rows, all fields': {
    
        'topic': function () {
            myTabler.selectMulti(
                myTablerName,
                null,
                [myRowId1, myRowId2],
                this.callback
            );
        },
        
        'and it succeeds and the data matches': function (inError, inRowWithIdHash) {
                            
            assert.isNull     (inError);
            
            var compObject = {};
            compObject[myRowId1] = Util.deepCopy(myRow1);
            compObject[myRowId1]['__id'] = myRowId1;
            compObject[myRowId2] = Util.deepCopy(myRow2);
            compObject[myRowId2]['__id'] = myRowId2;
            
            assert.deepEqual  (inRowWithIdHash, compObject);
        }
        
    }

})
.addBatch({
        
    'try to get two existing rows, no fields (existence)': {
    
        'topic': function () {
            myTabler.selectMulti(
                myTablerName,
                [], // No fields
                [myRowId1, myRowId2],
                this.callback
            );
        },
        
        'and it succeeds and there are no fields': function (inError, inRowWithIdHash) {
                            
            assert.isNull(inError);
            assert.isObject(inRowWithIdHash);
            assert.isObject(inRowWithIdHash[myRowId1]);
            assert.isObject(inRowWithIdHash[myRowId1]);
            
            var k,
                count;

            count = 0;
            for (k in inRowWithIdHash[myRowId1]) {
                if (k !== '__id') {
                    count++;
                }
            }
            assert.strictEqual(count, 0);

            count = 0;
            for (k in inRowWithIdHash[myRowId2]) {
                if (k !== '__id') {
                    count++;
                }
            }
            assert.strictEqual(count, 0);
            
        }
        
    }

})
.addBatch({
        
    'try to get one existing and one non-existing row, all fields': {
    
        'topic': function () {
            myTabler.selectMulti(
                myTablerName,
                null,
                [myRowId1, 'lul_row'],
                this.callback
            );
        },
        
        'and it succeeds and the data matches': function (inError, inRowWithIdHash) {
                            
            assert.isNull     (inError);
            
            var compObject = {};
            compObject[myRowId1] = Util.deepCopy(myRow1);
            compObject[myRowId1]['__id'] = myRowId1;
            
            assert.deepEqual  (inRowWithIdHash, compObject);
        }
        
    }

})
.addBatch({
        
    'try to get two existing rows, one valid field': {
    
        'topic': function () {
            myTabler.selectMulti(
                myTablerName,
                [myFieldName1],
                [myRowId1, myRowId2],
                this.callback
            );
        },
        
        'and it succeeds and the data matches': function (inError, inRowWithIdHash) {
                            
            assert.isNull     (inError);
            
            var compObject = {};
            compObject[myRowId1] = {};
            compObject[myRowId1][myFieldName1] = myRow1[myFieldName1];
            compObject[myRowId1]['__id'] = myRowId1;
            compObject[myRowId2] = {};
            compObject[myRowId2][myFieldName1] = myRow2[myFieldName1];
            compObject[myRowId2]['__id'] = myRowId2;
            
            assert.deepEqual  (inRowWithIdHash, compObject);
        }
        
    }

})
.addBatch({
        
    'try to get many rows (none existing) at once': {
    
        'topic': function () {
        
            var getIds = [];
            for (var i = 0; i < 50; i++) {
                getIds.push('id' + i);
            }
        
            myTabler.selectMulti(
                myTablerName,
                [],
                getIds,
                this.callback
            );
        },
        
        'and it succeeds and no results are present': function (inError, inRowWithIdHash) {

            var n = 0;
            for (var k in inRowWithIdHash) {
            }
            assert.strictEqual(n, 0);
            
        }
        
    }

})
.addBatch({
        
    'try to run a select query with a bogus field name': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                ['junk_field'],
                [], // Condition
                this.callback
            );
        },
        
        'and it fails': function (inError, inRowWithIdArray) {
                            
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'NON_EXISTENT_FIELD');
            
            
        }
        
    }

})
.addBatch({
        
    'try to run a select query with a bogus condition field name': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [], // No fields
                [ ['junk_field', '=', 'any_value'] ],
                this.callback
            );
        },
        
        'and it fails': function (inError, inRowWithIdArray) {
                            
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'NON_EXISTENT_FIELD');
            
            
        }
        
    }

})
.addBatch({
        
    'try to run a select query with a bogus condition operator': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [],
                [ [myFieldName1, '<>', 'any_value'] ], // It's != not <>
                this.callback
            );
        },
        
        'and it fails': function (inError, inRowWithIdArray) {
                            
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'INVALID_OPERATOR');
            
        }
        
    }

})
.addBatch({
        
    'try to run a select query with a condition on a long (partitioned) field': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [], // Field names
                [ [myFieldName2, '=', 'any_value'] ],
                this.callback
            );
        },
        
        'and it fails': function (inError, inRowWithIdArray) {
        
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'CANNOT_CONDITION_ON_PARTITIONED_FIELD');
            
        }
        
    }

})
.addBatch({
        
    'try to run a select query with a bogus order field': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [], // Field names
                [], // Condition
                {
                    orderBy: 'lul_field'
                },
                this.callback
            );
        },
        
        'and it fails': function (inError, inRowWithIdArray) {
                            
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'INVALID_ORDER_FIELD');
            
        }
        
    }

})
.addBatch({
        
    'try to run a select query with ordering on a long (partitioned) field': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [], // Field names
                [], // Condition
                {
                    orderBy: myFieldName2
                },
                this.callback
            );
        },
        
        'and it fails': function (inError, inRowWithIdArray) {
        
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'CANNOT_ORDER_ON_PARTITIONED_FIELD');
            
        }
        
    }

})
.addBatch({
        
    'try to run a select query with a bogus order direction': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [], // Field names
                [], // Conditions
                {
                    orderBy: myFieldName1,
                    orderDir: 'lul_dir'
                },
                this.callback
            );
        },
        
        'and it fails': function (inError, inRowWithIdArray) {
                            
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'INVALID_ORDER_DIRECTION');
            
        }
        
    }

})
.addBatch({
        
    'try to run a basic query': {
    
        'topic': function () {
        
            var cb = this.callback;
            
            myTabler.selectQuery(
                myTablerName,
                null, // All fields
                [], // Conditions
                {
                    orderBy: 'username',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it returns the correct data': function (inError, inRowWithIdArray) {
        
            assert.isNull(inError);
            assert.equal(inRowWithIdArray.length, 3);
            
            // The order of the results is indeterminate - just check that
            // each row exists in the set
            var fCheck = function (n) {
                var id = inRowWithIdArray[n].__id
                delete inRowWithIdArray[n].__id;
                
                
                if (id === myRowId1) {
                    assert.deepEqual(inRowWithIdArray[n], myRow1);
                } else if (id === myRowId2) {
                    assert.deepEqual(inRowWithIdArray[n], myRow2);
                } else if (id === myRowId3) {
                    assert.deepEqual(inRowWithIdArray[n], myRow3);
                } else {
                    // Fail the test - there is probably junk data from a previous run in the db
                    assert.equal(0,1);
                }
            };
            fCheck(0);
            fCheck(1);
            fCheck(2);
                                
        }
        
    }

})
.addBatch({
        
    'try to run a basic query selecting no fields (want row ids only)': {
    
        'topic': function () {
        
            var cb = this.callback;
            
            myTabler.selectQuery(
                myTablerName,
                [], // No fields
                [], // Conditions
                // Default options
                this.callback
            );
        },
        
        'and it returns no field data, just ids': function (inError, inRowWithIdArray) {
            
            assert.isNull(inError);
            assert.equal(inRowWithIdArray.length, 3);
            
            var i,
                k,
                count;
            for (i = 0; i < inRowWithIdArray.length; i++) {
                count = 0;
                for (k in inRowWithIdArray[i]) {
                    if (k !== '__id') {
                        count++;
                    }
                }
                assert.strictEqual(count, 0);
            }
              
        }
        
    }

})
.addBatch({
        
    'try to run a basic query with a result limit': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [],
                [],
                {
                    limit: 2
                },
                this.callback
            );
        },
        
        'and the correct number of results are returned': function (inError, inRowWithIdArray) {
                                
            assert.isNull(inError);
            assert.equal(inRowWithIdArray.length, 2);
                                
        }
        
    }

})
.addBatch({
        
    'try to run a basic query with an order field': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                null, // All fields
                [],
                {
                    orderBy: myFieldName1
                },
                this.callback
            );
        },
        
        'and the correct results are returned': function (inError, inRowWithIdArray) {
            
            assert.isNull(inError);
            
            assert.equal(inRowWithIdArray.length, 3);
            
            var expectedArray = [];
            expectedArray[0] = myRow1;
            expectedArray[0].__id = myRowId1;
            expectedArray[1] = myRow2;
            expectedArray[1].__id = myRowId2;
            expectedArray[2] = myRow3;
            expectedArray[2].__id = myRowId3;
            
            assert.deepEqual (expectedArray, inRowWithIdArray);
                                
        }
        
    }

})
.addBatch({
        
    'try to run a query with an order field, descending': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                null, // All fields
                [],
                {
                    orderBy: myFieldName1,
                    orderDir: 'desc',
                    limit: 0
                },
                this.callback
            );
        },
        
        'and the correct results are returned': function (inError, inRowWithIdArray) {
            
            assert.isNull(inError);
            
            assert.equal(inRowWithIdArray.length, 3);
            
            var expectedArray = new Array(3);
            expectedArray[2] = myRow1;
            expectedArray[2].__id = myRowId1;
            expectedArray[1] = myRow2;
            expectedArray[1].__id = myRowId2;
            expectedArray[0] = myRow3;
            expectedArray[0].__id = myRowId3;
            
            assert.deepEqual (expectedArray, inRowWithIdArray);
                                
        }
        
    }

})
.addBatch({
        
    'try to run a query with a field list': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [myFieldName1],
                [],
                this.callback
            );
        },
        
        'and the correct results are returned': function (inError, inRowWithIdArray) {
            
            assert.isNull(inError);
                                
            for (var i = 0; i < inRowWithIdArray.length; i++) {
                for (var k in inRowWithIdArray[i]) {
                    assert.isTrue(k === '__id' || k === myFieldName1);
                }
            }
                                
        }
        
    }

})
.addBatch({
        
    'try to run a query with one condition': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                null, // All fields
                [ [myFieldName1, '=', myRow1[myFieldName1]] ],
                this.callback
            );
        },
        
        'and the correct results are returned': function (inError, inRowWithIdArray) {
        
            assert.isNull(inError);
            
            var expectedArray = [];
            expectedArray[0] = myRow1;
            expectedArray[0].__id = myRowId1;
                                
            assert.deepEqual(expectedArray, inRowWithIdArray);
                                
        }
        
    }

})
.addBatch({
        
    'try to run a query with two conditions': {
    
        'topic': function () {
            myTabler.selectQuery(
                myTablerName,
                [],
                [
                    [myFieldName1, '=', myRow1[myFieldName1]],
                    [myFieldName1, '=', 'some_junk']
                ],
                this.callback
            );
        },
        
        'and the correct results are returned': function (inError, inRowWithIdArray) {
            
            assert.isNull(inError);
            assert.strictEqual(inRowWithIdArray.length, 0);
                                
        }
        
    }

})
.addBatch({
        
    'try to run one update with a non-existent field name': {
    
        'topic': function () {
            myTabler.updateOne(
                myTablerName,
                {
                    'junk_field': 'any_value'
                },
                myRowId1,
                this.callback
            );
        },
        
        'and an error is returned': function (inError, inRowWithIdArray) {
            
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'EXTRANEOUS_FIELD');
            
        }
        
    }

})
.addBatch({
        
    'try to run one update with a non-string value': {
    
        'topic': function () {
            var arg = {};
            arg[myFieldName1] = 1;
            
            myTabler.updateOne(
                myTablerName,
                arg,
                myRowId1,
                this.callback
            );
        },
        
        'and an error is returned': function (inError, inRowWithIdArray) {
            
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'VALUE_NOT_STRING');
            
        }
        
    }

})
.addBatch({
        
    'try to run one update with a too-long value': {
    
        'topic': function () {
            var arg = {};
            arg[myFieldName1] = (new Array(1000000)).join('j');
            
            myTabler.updateOne(
                myTablerName,
                arg,
                myRowId1,
                this.callback
            );
        },
        
        'and an error is returned': function (inError, inRowWithIdArray) {
            
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'VALUE_TOO_LONG');
            
        }
        
    }

})
.addBatch({
        
    'try to run one update on a non-existent row': {
    
        'topic': function () {
            var arg = {};
            arg[myFieldName1] = 'hi';
            
            myTabler.updateOne(
                myTablerName,
                arg,
                'some_junk_row_id',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            
            // The value is NOT written - tested below
            
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'try to load the update on the non-existent row': {
    
        'topic': function () {
            
            myTabler.selectOne(
                myTablerName,
                [],
                'some_junk_row_id', // Same as above
                this.callback
            );
        },
        
        'and it succeeds and the row is not found': function (inError, inRowWithoutId) {
                            
            assert.equal(inError, null);
            assert.equal(inRowWithoutId, null);
            
        }
        
    }

})
.addBatch({
        
    'try to run one update on a valid row': {
    
        'topic': function () {
            var arg = {};
            arg[myFieldName1] = 'hi';
            
            myTabler.updateOne(
                myTablerName,
                arg,
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            
            // The value IS written -- tested next
            
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'try to load the update on the valid row': {
    
        'topic': function () {
            
            myTabler.selectOne(
                myTablerName,
                null, // All fields
                myRowId1, // Same as above
                this.callback
            );
        },
        
        'and it succeeds and the update is present': function (inError, inRowWithoutId) {
                            
            assert.equal(inError, null);
            assert.isObject(inRowWithoutId);
            assert.equal(inRowWithoutId[myFieldName1], 'hi');
            assert.equal(inRowWithoutId[myFieldName2], myRow1[myFieldName2]);
            
        }
        
    }

})
.addBatch({
        
    'try to apply a large (partitioned) update to one row': {
    
        'topic': function () {
            var arg = {};
            arg[myFieldName2] = (new Array(1501)).join('x'); // > 1024
            
            myTabler.updateOne(
                myTablerName,
                arg,
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            
            // The value IS written -- tested next
            
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'try to load the long-value update': {
    
        'topic': function () {
            
            myTabler.selectOne(
                myTablerName,
                null, // All fields
                myRowId1, // Same as above
                this.callback
            );
        },
        
        'and it succeeds and the update is present': function (inError, inRowWithoutId) {
                            
            assert.equal(inError, null);
            assert.isObject(inRowWithoutId);
            assert.equal(inRowWithoutId[myFieldName1], 'hi');
            assert.equal(inRowWithoutId[myFieldName2], (new Array(1501)).join('x'));
            
        }
        
    }

})
.addBatch({
        
    'try to apply a small (empty) value over the large (partitioned) value for one row': {
    
        'topic': function () {
        
            var arg = {};
            arg[myFieldName2] = 'a'; // < 1024
            
            myTabler.updateOne(
                myTablerName,
                arg,
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            
            // The value IS written -- tested next
            
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'try to load the short-value update': {
    
        'topic': function () {
            
            myTabler.selectOne(
                myTablerName,
                null, // All fields
                myRowId1, // Same as above
                this.callback
            );
        },
        
        'and it succeeds and the update is present': function (inError, inRowWithoutId) {
        
            assert.equal(inError, null);
            assert.isObject(inRowWithoutId);
            assert.equal(inRowWithoutId[myFieldName1], 'hi');
            assert.equal(inRowWithoutId[myFieldName2], 'a');
            
        }
        
    }

})
.addBatch({
        
    'try to update one row using a query': {
    
        'topic': function () {
            var arg = {};
            arg[myFieldName1] = 'Johnnay';
            
            myTabler.updateQuery(
                myTablerName,
                arg,
                [ ['username', '=', 'hi' ] ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that the update took place properly': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                [],
                [ [myFieldName1, '=', 'Johnnay'] ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, 1);
            
        }
        
    }

})
.addBatch({
        
    'try to update all rows using a query': {
    
        'topic': function () {
            var arg = {};
            arg[myFieldName1] = 'some_username';
            
            myTabler.updateQuery(
                myTablerName,
                arg,
                [], // Entire table
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that the update took place properly': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                null, // All fields
                [ [myFieldName1, '=', 'some_username'] ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
                            
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, 3);
            assert.strictEqual(inRowsWithIdArray[0].username, 'some_username');
            assert.strictEqual(inRowsWithIdArray[1].username, 'some_username');
            assert.strictEqual(inRowsWithIdArray[2].username, 'some_username');
            
        }
        
    }

})
.addBatch({
        
    'try to delete one row': {
    
        'topic': function () {
        
            myTabler.deleteOne(
                myTablerName,
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that the delete took place properly': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                [],
                [],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
                            
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, 2);
            
        }
        
    }

})
.addBatch({
        
    'try to delete two rows, one exists': {
    
        'topic': function () {
        
            myTabler.deleteMulti(
                myTablerName,
                [myRowId1, myRowId2],
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {

            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that the delete took place properly': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                [],
                [],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
                            
            assert.equal(inError, null);
            
            assert.strictEqual(inRowsWithIdArray.length, 1);
            
        }
        
    }

})
.addBatch({
        
    'run a delete query that is not satisfied by any rows': {
    
        'topic': function () {
        
            myTabler.deleteQuery(
                myTablerName,
                [ [myFieldName1, '=', 'any_value'] ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {

            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that nothing happened': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                [],
                [],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
                            
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, 1);
            
        }
        
    }

})
.addBatch({
        
    'run a delete query to empty the table': {
    
        'topic': function () {
        
            myTabler.deleteQuery(
                myTablerName,
                [],
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {

            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that the table is empty': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                [],
                [],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
                            
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, 0);
            
        }
        
    }

})
.addBatch({
        
    'run a delete query that empties the table': {
    
        'topic': function () {
        
            myTabler.deleteQuery(
                myTablerName,
                [],
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {

            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that the table is empty': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                [],
                [],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
                            
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, 0);
            
        }
        
    }

})
.addBatch({
        
    'table should be empty, try to insert many rows': {
    
        'topic': function () {
        
            var ins = [],
                curUser;
            for (var i = 0; i < myBigNum; i++) {
            
                curUser = '' + i;
                while (curUser.length < 10) {
                    curUser = '0' + curUser;
                }
                ins.push({
                    username:   curUser,
                    post:       'hi'
                });
            }
            
            myTabler.insertMulti(
                myTablerName,
                ins,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.isNull     (inError);
                        
        }
        
    }

})
.addBatch({
        
    'check that the table contains the rows and that they can be sorted properly': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                null, // All fields
                [], // No conditions
                {
                    orderBy: 'username',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, myBigNum);
            
            var expUser;
            for (var i = myBigNum - 1; i >= 0; i--) {
                expUser = '' + i;
                while (expUser.length < 10) {
                    expUser = '0' + expUser;
                }
                assert.strictEqual(inRowsWithIdArray[myBigNum - i - 1].username, expUser);
            }
            
        }
        
    }

})
.addBatch({
        
    'select a portion of the data': {
    
        'topic': function () {
        
            var num = '' + Math.floor(myBigNum / 2);
            while (num.length < 10) {
                num = '0' + num;
            }
            
            myTabler.selectQuery(
                myTablerName,
                ['username'],
                [ ['username', '<', num] ],
                {
                    orderBy: 'username',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowsWithIdArray) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, Math.floor(myBigNum / 2));
                                
            var expUser;
            for (var i = Math.floor(myBigNum / 2) - 1; i >= 0; i--) {
                expUser = '' + i;
                while (expUser.length < 10) {
                    expUser = '0' + expUser;
                }
                assert.strictEqual(inRowsWithIdArray[Math.floor(myBigNum / 2) - i - 1].username, expUser);
            }
            
        }
        
    }

})
.addBatch({
        
    'update a portion of the data': {
    
        'topic': function () {
        
            var num = '' + Math.floor(myBigNum / 2);
            while (num.length < 10) {
                num = '0' + num;
            }
            
            myTabler.updateQuery(
                myTablerName,
                {
                    username: 'new_username'
                },
                [ ['username', '<', num] ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            // Check write is next step
            
        }
        
    }

})
.addBatch({
        
    'check that the update took place successfully': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                ['username'],
                [ ['username', '=', 'new_username'] ],
                this.callback
            );
        },
        
        'took place successfully': function (inError, inRowsWithIdArray) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, Math.floor(myBigNum / 2));
                            
        }
        
    }

})
.addBatch({
        
    'delete a portion of the data': {
    
        'topic': function () {
            
            myTabler.deleteQuery(
                myTablerName,
                [ ['username', '=', 'new_username'] ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            // Check write is next step
            
        }
        
    }

})
.addBatch({
        
    'check that the delete took place successfully': {
    
        'topic': function () {
        
            myTabler.selectQuery(
                myTablerName,
                [],
                [],
                this.callback
            );
        },
        
        'took place successfully': function (inError, inRowsWithIdArray) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRowsWithIdArray.length, myBigNum - Math.floor(myBigNum / 2));
                            
        }
        
    }

})
.addBatch({
        
    'run an update with a sketchy condition': {
    
        'topic': function () {
        
            myTabler.updateQuery(
                myTablerName,
                {
                    'username': 'no_chango'
                },
                'bad_condition',
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_CONDITION');
                            
        }
        
    }

})
.addBatch({
        
    'run an update with another sketchy condition': {
    
        'topic': function () {
        
            myTabler.updateQuery(
                myTablerName,
                {
                    'username': 'no_chango'
                },
                ['username', '=', 'hi'], // Not a valid condition
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_CONDITION');
                            
        }
        
    }

})
.addBatch({
        
    'run an update with another sketchy condition': {
    
        'topic': function () {
        
            myTabler.updateQuery(
                myTablerName,
                {
                    'username': 'no_chango'
                },
                [['username', '=', 'hi', '2']], // Not a valid condition
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_CONDITION');
                            
        }
        
    }

})
.addBatch({
        
    'run a delete with a sketchy condition': {
    
        'topic': function () {
        
            myTabler.deleteQuery(
                myTablerName,
                'bad_condition',
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_CONDITION');
                            
        }
        
    }

})
.addBatch({
        
    'run a delete with another sketchy condition': {
    
        'topic': function () {
        
            myTabler.deleteQuery(
                myTablerName,
                ['username', '=', 'hi'], // Not a valid condition
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_CONDITION');
                            
        }
        
    }

})
.addBatch({
        
    'run a delete with another sketchy condition': {
    
        'topic': function () {
        
            myTabler.deleteQuery(
                myTablerName,
                [['username', '=', 'hi', '2']], // Not a valid condition
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_CONDITION');
                            
        }
        
    }

})
.addBatch({
        
    'try to insert a string that is too long by bytes but NOT by characters': {
    
        'topic': function () {
        
            myTabler.insertOne(
                myTablerName,
                {
                    'username': 'ÚÚÚÚÚÚÚÚÚÚÚ', // 2 bytes per character
                    'post': ''
                },
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'VALUE_TOO_LONG');
                            
        }
        
    }

})
.addBatch({
        
    'try to update a string that is too long by bytes but NOT by characters': {
    
        'topic': function () {
        
            myTabler.updateQuery(
                myTablerName,
                {
                    'username': 'ÚÚÚÚÚÚÚÚÚÚÚ', // 2 bytes per character
                },
                [], // All (will fail)
                this.callback
            );
        },
        
        'and it failed': function (inError, inRowsWithIdArray) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'VALUE_TOO_LONG');
                            
        }
        
    }

})
.addBatch({
        
    'try to get the number of rows in the table': {
    
        'topic': function () {
        
            myTabler.countQuery(
                myTablerName,
                [], // All
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inCount) {
        
            assert.isNull(inError);
            assert.isNumber(inCount);
            assert.strictEqual(inCount, myBigNum - Math.floor(myBigNum / 2));
            
        }
        
    }

})


// Test specifying an id when inserting a row
.addBatch({
        
    'clear the table': {
    
        'topic': function () {
        
            myTabler.deleteQuery(
                myTablerName,
                [], // All
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.isNull(inError);
            
        }
        
    }

})
.addBatch({
        
    'insert a row with an id specifiied': {
    
        'topic': function () {
        
            myTabler.insertOne(
                myTablerName,
                {
                    __id:       'testid',
                    username:   'aaron',
                    post:       'mypost'
                    
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.isNull(inError);
            
        }
        
    }

})
.addBatch({
        
    'check that the row was inserted successfully with the correct id': {
    
        'topic': function () {
        
            myTabler.selectOne(
                myTablerName,
                null, // all
                'testid',
                this.callback
            );
        },
        
        'and it was': function (inError, inRowWithoutId) {

            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'try to insert a row with a duplicate id specifiied': {
    
        'topic': function () {
        
            myTabler.insertOne(
                myTablerName,
                {
                    __id:       'testid',
                    username:   'aaron',
                    post:       'mypost'
                    
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'ID_EXISTS');
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two rows with a duplicate id specified': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                myTablerName,
                [
                {
                    __id:       'testid',
                    username:   'aaron',
                    post:       'mypost'
                    
                },
                {
                    __id:       'new_id',
                    username:   'aaron',
                    post:       'mypost'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'ID_EXISTS');
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two rows with a duplicate id specified in the new rows': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                myTablerName,
                [
                {
                    __id:       'new_id',
                    username:   'aaron',
                    post:       'mypost'
                    
                },
                {
                    __id:       'new_id',
                    username:   'aaron',
                    post:       'mypost'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'DUPLICATE_IDS');
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two valid rows with specified ids': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                myTablerName,
                [
                {
                    __id:       'new_id',
                    username:   'aaron',
                    post:       'mypost'
                    
                },
                {
                    __id:       'new_id2',
                    username:   'aaron',
                    post:       'mypost'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inIds) {
        
            assert.isNull(inError);
            assert.deepEqual(inIds, ['new_id', 'new_id2']);
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two valid rows with one id specified': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                myTablerName,
                [
                {
                    __id:       'new_id3',
                    username:   'aaron',
                    post:       'mypost'
                    
                },
                {
                    username:   'aaron',
                    post:       'mypost'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inIds) {
        
            assert.isNull(inError);
            assert.deepEqual(inIds[0], 'new_id3');
            assert.isString(inIds[1]); // Some GUID
            
        }
        
    }

})





// Table deletion 
.addBatch({
        
    'try to delete a junk table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'some_junk_table',
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'NON_EXISTENT_TABLE')
                        
        }
        
    }

})
.addBatch({
        
    'try to delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                myTablerName,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.isNull(inError);
                        
        }
        
    }

})
.addBatch({
        
    'try to access the deleted table': {
    
        'topic': function () {
            myTabler.selectOne(
                myTablerName,
                [],
                'a',
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'NON_EXISTENT_TABLE');
                        
        }
        
    }

})
.run({
    error: false // Do not auto-detect the first argument of vows as an error, I check these myself
}, function () {
    
    if (process.argv[2] === 'jsonfile') {
        fs.unlinkSync('stringtable.json');
    }
    
});


console.log(
    '-----\nRemember that if test fail then they might leave data, in which case if you run '
    + 'them again they could fail. Always try twice.\n-----'
);

