﻿// You pass a command line argument of the file that the
// module is in
if (process.argv.length < 3) {
    console.log('Run something like: node tabler.js jsonfile');
    process.exit();
}
if (process.argv[2] === 'simpledb' && process.argv.length !== 5) {
    console.log('Run something like: node tabler.js simpledb keyid key');
    process.exit();
}




// Module
var vows = require('vows'),
    assert = require('assert'),
    Tabler = require('../lib/tabler'),
    Util = require('../lib/util'),
    fs = require('fs');
    


    


var suite = vows.describe('Tabler');


/*

How Vows works:

    Batches are run in series
    Contexts in a batch are run in parallel
    A context's topic returns what you are testing
    A context's vows are the tests run on the topic
    
    
    
First test that all the functions work with strings, then ensure
that all the data types work.
    
    
*/

var myTabler,
    myRowId1,
    myRowId2,
    myRowId3,
    myCurDate; // Changes
    
suite.addBatch({
    
    'instantiate the object': {
    
        'topic': function () {
        
            if (process.argv[2] === 'simpledb') {
            
                return myTabler = new Tabler(
                    'simpledb',
                    {
                        accessKeyId:        process.argv[3],
                        secretAccessKey:    process.argv[4],
                        domainNamespace:    'testtype'
                    }
                );
            
            } else {
            
                return myTabler = new Tabler(
                    'jsonfile',
                    {
                        filename: 'test.json'
                    }
                );
                
            }
            
            
        },
        
        'worked': function (inTypeTable) {
            assert.isObject(myTabler);
        }
        
    }
        
})
.addBatch({
    
    'try to define an invalid table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':          'invalid'
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    'try to define another invalid table (bad string params)': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':         {type: 'string', pattern: /.*/} // Missing byteLength
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    'try to define another invalid table (bad string params)': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':         {type: 'string', pattern: /.*/, byteLength: "1"}
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    'try to define a valid table (string)': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'username': {
                        type: 'string',
                        pattern: myTabler.PATTERNS.PRINTABLE_NO_BREAKS,
                        byteLength: 10
                    }
                },
                this.callback
            );
        },
        
        'and it works': function (inError) {
        
            assert.equal(inError, null); // Not strict
        }
  
    }
    
})
.addBatch({
        
    'try to define a duplicate table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'username': {
                        type: 'string',
                        pattern: /^[^q]*$/,
                        byteLength: 10
                    }
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'TABLE_ALREADY_DEFINED');
        }
        
    }

})
.addBatch({
        
    'try to select one non-existent row': {
    
        'topic': function () {
            myTabler.selectOne(
                'myTabler',
                [],
                'non-ex',
                this.callback
            );
        },
        
        'and it returns null': function (inError, inRowWithoutId) {
            assert.equal(inError, null);
            assert.isNull(inRowWithoutId);
        }
        
    }

})
.addBatch({
        
    'try to select two non-existent rows': {
    
        'topic': function () {
            myTabler.selectMulti(
                'myTabler',
                [],
                ['a', 'b'],
                this.callback
            );
        },
        
        'and it returns empty object': function (inError, inResult) {

            assert.equal(inError, null);
            assert.deepEqual(inResult, {});
        }
        
    }

})


.addBatch({
        
    'try to insert one row with an invalid value (string)': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'username': 1
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'username');
        }
        
    }

})
.addBatch({
        
    'try to insert one row with an invalid value (string)': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'username': 'ÚÚÚÚÚÚ' // 11 bytes
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'username');
        }
        
    }

})
.addBatch({
        
    'try to insert another row with an invalid value (string)': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'username': 'haq\u0000' // Violates the pattern
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'username');
        }
        
    }

})
.addBatch({
        
    'try to insert one row with a good value (string)': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'username': "ÚÚÚÚÚ"
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowId) {
            assert.equal(null, inError);
            myRowId1 = inRowId;
        }
        
    }

})
.addBatch({
        
    'try to insert multiple rows with an invalid value (string)': {
    
        'topic': function () {
            myTabler.insertMulti(
                'myTabler',
                [
                    {
                        'username': 'okvalue'
                    },
                    {
                        'username': 'ÚÚÚÚÚÚ' // 12 bytes
                    }
                ],
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'username');
        }
        
    }

})
.addBatch({
        
    'try to insert multiple rows with good values (string)': {
    
        'topic': function () {
            myTabler.insertMulti(
                'myTabler',
                [
                    {
                        'username': 'okvalue'
                    },
                    {
                        'username': '1ÚÚÚÚ'
                    }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowIdArray) {
            assert.equal(null, inError);
            myRowId2 = inRowIdArray[0];
            myRowId3 = inRowIdArray[1];
        }
        
    }

})
.addBatch({
        
    'try to retreive one row (string)': {
    
        'topic': function () {
            myTabler.selectOne(
                'myTabler',
                null, // All fields
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds and the data is correct': function (inError, inRow) {
        
            assert.equal(null, inError);
            assert.strictEqual(inRow.username, "ÚÚÚÚÚ");
            
        }
        
    }

})
.addBatch({
        
    'try to retreive two rows (string)': {
    
        'topic': function () {
            myTabler.selectMulti(
                'myTabler',
                null, // All fields
                [myRowId2, myRowId3],
                this.callback
            );
        },
        
        'and it succeeds and the data is correct': function (inError, inRows) {
        
            assert.equal(null, inError);
            assert.strictEqual(inRows[myRowId2].username, "okvalue");
            assert.strictEqual(inRows[myRowId3].username, "1ÚÚÚÚ");
            
        }
        
    }

})
.addBatch({
        
    'try to retreive using a query (string)': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All fields
                [
                    ['username', '=', '1ÚÚÚÚ']
                ],
                this.callback
            );
        },
        
        'and it succeeds and the data is correct': function (inError, inRows) {
        
            assert.equal(null, inError);
            assert.strictEqual(inRows.length, 1);
            assert.strictEqual(inRows[0].username, "1ÚÚÚÚ");
            
        }
        
    }

})
.addBatch({
        
    'try to count (string)': {
    
        'topic': function () {
            myTabler.countQuery(
                'myTabler',
                [
                    ['username', '=', '1ÚÚÚÚ']
                ],
                this.callback
            );
        },
        
        'and it succeeds and the data is correct': function (inError, inCount) {
        
            assert.equal(null, inError);
            assert.strictEqual(inCount, 1);
            
        }
        
    }

})
.addBatch({
        
    'try to update one (string)': {
    
        'topic': function () {
            myTabler.updateOne(
                'myTabler',
                {
                    username: 'new_user'
                },
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds (data tested next)': function (inError) {
        
            assert.equal(null, inError);
            
        }
        
    }

})
.addBatch({
        
    'check that the update occurred correctly': {
    
        'topic': function () {
            myTabler.selectOne(
                'myTabler',
                null, // All fields
                myRowId1,
                this.callback
            );
        },
        
        'and it did': function (inError, inRow) {
        
            assert.equal(null, inError);
            assert.strictEqual(inRow.username, 'new_user');
            
        }
        
    }

})
.addBatch({
        
    'try to update multiple (string)': {
    
        'topic': function () {
            myTabler.updateMulti(
                'myTabler',
                {
                    username: 'even_newer'
                },
                [myRowId2, myRowId3],
                this.callback
            );
        },
        
        'and it succeeds (data tested next)': function (inError) {
        
            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'check that the update occurred correctly': {
    
        'topic': function () {
            myTabler.selectMulti(
                'myTabler',
                null, // All fields
                [myRowId2, myRowId3],
                this.callback
            );
        },
        
        'and it did': function (inError, inRowWithIdHash) {
        
            assert.equal(null, inError);
            assert.strictEqual(inRowWithIdHash[myRowId2].username, 'even_newer');
            assert.strictEqual(inRowWithIdHash[myRowId3].username, 'even_newer');
            
        }
        
    }

})
.addBatch({
        
    'try to run an update query (string)': {
    
        'topic': function () {
            myTabler.updateQuery(
                'myTabler',
                {
                    username: 'newest_yet'
                },
                [
                    ['username', '=', 'even_newer']
                ],
                this.callback
            );
        },
        
        'and it succeeds (data tested next)': function (inError) {
        
            assert.equal(null, inError);
            
        }
        
    }

})
.addBatch({
        
    'check that the update occurred correctly': {
    
        'topic': function () {
            myTabler.selectMulti(
                'myTabler',
                null, // All fields
                [myRowId2, myRowId3],
                this.callback
            );
        },
        
        'and it did': function (inError, inRowWithIdHash) {
        
            assert.equal(null, inError);
            assert.strictEqual(inRowWithIdHash[myRowId2].username, 'newest_yet');
            assert.strictEqual(inRowWithIdHash[myRowId3].username, 'newest_yet');
            
        }
        
    }

})
.addBatch({
        
    'try to run delete one row': {
    
        'topic': function () {
            myTabler.deleteOne(
                'myTabler',
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds (data tested next)': function (inError) {
        
            assert.equal(null, inError);
            
        }
        
    }

})
.addBatch({
        
    'check that the delete occurred correctly': {
    
        'topic': function () {
            myTabler.countQuery(
                'myTabler',
                [], // No condition
                this.callback
            );
        },
        
        'and it did': function (inError, inCount) {
        
            assert.equal(null, inError);
            assert.strictEqual(inCount, 2);
            
        }
        
    }

})
.addBatch({
        
    'try to run delete multiple rows': {
    
        'topic': function () {
            myTabler.deleteMulti(
                'myTabler',
                [myRowId1, myRowId2], // One is already done
                this.callback
            );
        },
        
        'and it succeeds (data tested next)': function (inError) {
        
            assert.equal(null, inError);
            
        }
        
    }

})
.addBatch({
        
    'check that the delete occurred correctly': {
    
        'topic': function () {
            myTabler.countQuery(
                'myTabler',
                [], // No condition
                this.callback
            );
        },
        
        'and it did': function (inError, inCount) {
        
            assert.equal(null, inError);
            assert.strictEqual(inCount, 1);
            
        }
        
    }

})
.addBatch({
        
    'try to run delete using a query': {
    
        'topic': function () {
            myTabler.deleteQuery(
                'myTabler',
                [
                    ['username', '!=', 'hola'] // All
                ],
                this.callback
            );
        },
        
        'and it succeeds (data tested next)': function (inError) {
        
            assert.equal(null, inError);
            
        }
        
    }

})
.addBatch({
        
    'check that the delete occurred correctly': {
    
        'topic': function () {
            myTabler.countQuery(
                'myTabler',
                [], // No condition
                this.callback
            );
        },
        
        'and it did': function (inError, inCount) {
        
            assert.equal(null, inError);
            assert.strictEqual(inCount, 0);
            
        }
        
    }

})
.addBatch({
        
    'try to delete a junk table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'some_junk_table',
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'NON_EXISTENT_TABLE')
                        
        }
        
    }

})
.addBatch({
        
    'try to delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})
.addBatch({
        
    'try to access the deleted table': {
    
        'topic': function () {
            myTabler.selectOne(
                'myTabler',
                [],
                'some_id',
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject     (inError);
            assert.strictEqual  (inError.code, 'NON_EXISTENT_TABLE');
                        
        }
        
    }

})

// Test specifying an id when inserting a row
.addBatch({
    
    'try to define a valid table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'username': {
                        type: 'string',
                        byteLength: 100,
                        pattern: /.*/
                    }
                },
                this.callback
            );
        },
        
        'and it works': function (inError) {
            assert.equal(inError, null); // Not strict
        }
  
    }
    
})
.addBatch({
        
    'insert a row with an id specifiied': {
    
        'topic': function () {
        
            myTabler.insertOne(
                'myTabler',
                {
                    __id:       'testid',
                    username:   'aaron'
                    
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.isNull(inError);
            
        }
        
    }

})
.addBatch({
        
    'check that the row was inserted successfully with the correct id': {
    
        'topic': function () {
        
            myTabler.selectOne(
                'myTabler',
                null, // all
                'testid',
                this.callback
            );
        },
        
        'and it was': function (inError, inRowWithoutId) {

            assert.equal(inError, null);
            
        }
        
    }

})
.addBatch({
        
    'try to insert a row with a duplicate id specifiied': {
    
        'topic': function () {
        
            myTabler.insertOne(
                'myTabler',
                {
                    __id:       'testid',
                    username:   'aaron'
                    
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'ID_EXISTS');
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two rows with a duplicate id specified': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                'myTabler',
                [
                {
                    __id:       'testid',
                    username:   'aaron'
                    
                },
                {
                    __id:       'new_id',
                    username:   'aaron'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'ID_EXISTS');
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two rows with a duplicate id specified in the new rows': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                'myTabler',
                [
                {
                    __id:       'new_id',
                    username:   'aaron'
                    
                },
                {
                    __id:       'new_id',
                    username:   'aaron'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it fails': function (inError) {
        
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'DUPLICATE_IDS');
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two valid rows with specified ids': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                'myTabler',
                [
                {
                    __id:       'new_id',
                    username:   'aaron'
                    
                },
                {
                    __id:       'new_id2',
                    username:   'aaron'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inIds) {
        
            assert.isNull(inError);
            assert.deepEqual(inIds, ['new_id', 'new_id2']);
            
        }
        
    }

})
.addBatch({
        
    'try to insert a two valid rows with one id specified': {
    
        'topic': function () {
        
            myTabler.insertMulti(
                'myTabler',
                [
                {
                    __id:       'new_id3',
                    username:   'aaron'
                    
                },
                {
                    username:   'aaron'
                    
                }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inIds) {
        
            assert.isNull(inError);
            assert.deepEqual(inIds[0], 'new_id3');
            assert.isString(inIds[1]); // Some GUID
            
        }
        
    }

})
.addBatch({
        
    'delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})


// Null testing
.addBatch({
    
    '(null) try to define a table with a field that does not allow nulls': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'myField':         {type: 'string', byteLength: 50, pattern: /.*/} // Default is nullable:falsy
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            assert.equal(inError, null);
        }
  
    }
})
.addBatch({
    
    '(null) try to insert a null value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': null
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'NOT_NULLABLE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(null) try to covertly insert a null value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': "\u0000"
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'NOT_NULLABLE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
        
    '(null) delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})
.addBatch({
    
    '(null) try to define a table with a field that does allow nulls': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'myField':         {type: 'number', min: -100, max: 100, decPlaces: 0, nullable: true}
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            assert.equal(inError, null);
        }
  
    }
})
.addBatch({
    
    '(null) try to insert a null value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': null
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowId) {
            assert.equal(inError, null);
            myRowId1 = inRowId;
        }
  
    }
})
.addBatch({
    
    '(null) try to load the null value': {
    
        'topic': function () {
            myTabler.selectOne(
                'myTabler',
                null, // All
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRow) {
        
            assert.equal(inError, null);
            assert.deepEqual(inRow, {myField: null});
            
        }
  
    }
})
.addBatch({
    
    '(null) insert two more valid values': {
    
        'topic': function () {
            myTabler.insertMulti(
                'myTabler',
                [
                    {
                        'myField': 9
                    },
                    {
                        'myField': -15
                    }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowIds) {
        
            assert.equal(inError, null);
            myRowId2 = inRowIds[0];
            myRowId3 = inRowIds[1];
        }
  
    }
})
.addBatch({
    
    '(null) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(
                inRows,
                [{myField: 9, __id: myRowId2}, {myField: -15, __id: myRowId3}, {myField: null, __id: myRowId1}]
            );
            
        }
  
    }
})
.addBatch({
    
    '(null) load the values with opposite ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'asc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(
                inRows,
                [{myField: null, __id: myRowId1}, {myField: -15, __id: myRowId3}, {myField: 9, __id: myRowId2}]
                );
            
        }
  
    }
})
.addBatch({
    
    '(null) update a value': {
    
        'topic': function () {
            myTabler.updateOne(
                'myTabler',
                {
                    'myField': null
                },
                myRowId2,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
  
    }
})
.addBatch({
    
    '(null) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRows[0].__id, myRowId3);
            // Null rows could be in either order
            assert.strictEqual(inRows[0].myField, -15);
            assert.strictEqual(inRows[1].myField, null);
            assert.strictEqual(inRows[2].myField, null);
            
        }
  
    }
})
.addBatch({
        
    '(null) delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})


// Number tests (int)
.addBatch({
    
    '(number.int) try to define a table with an invalid type': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':         {type: 'number'} // Missing min, max, decPlaces
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    '(number.int) try to define another table with an invalid type': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':         {type: 'number', min: '1', max: '2', decPlaces: '2'}
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    '(number.int) try to define valid table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'number':         {type: 'number', min: -100, max: 100, decPlaces: 0}
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            assert.equal(inError, null);
        }
  
    }
})
.addBatch({
    
    '(number.int) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'number': 'hi'
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'number');
        }
  
    }
})
.addBatch({
    
    '(number.int) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'number': -101
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'number');
        }
  
    }
})
.addBatch({
    
    '(number.int) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'number': 101
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'number');
        }
  
    }
})
.addBatch({
    
    '(number.int) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'number': 0.1
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'number');
        }
  
    }
})
.addBatch({
    
    '(number.int) insert a valid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'number': 10
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowId) {
        
            assert.equal(inError, null);
            myRowId1 = inRowId;
        }
  
    }
})
.addBatch({
    
    '(number.int) insert two more valid values': {
    
        'topic': function () {
            myTabler.insertMulti(
                'myTabler',
                [
                    {
                        'number': 9
                    },
                    {
                        'number': -15
                    }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowIds) {
        
            assert.equal(inError, null);
            myRowId2 = inRowIds[0];
            myRowId3 = inRowIds[1];
        }
  
    }
})
.addBatch({
    
    '(number.int) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'number',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(
                inRows,
                [{number: 10, __id: myRowId1}, {number: 9, __id: myRowId2}, {number: -15, __id: myRowId3}]
            );
            
        }
  
    }
})
.addBatch({
    
    '(number.int) load the values with opposite ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'number',
                    orderDir: 'asc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(
                inRows,
                [{number: -15, __id: myRowId3}, {number: 9, __id: myRowId2}, {number: 10, __id: myRowId1}]
            );
            
        }
  
    }
})
.addBatch({
    
    '(number.int) update a value': {
    
        'topic': function () {
            myTabler.updateOne(
                'myTabler',
                {
                    'number': -4
                },
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
  
    }
})
.addBatch({
    
    '(number.int) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'number',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(
                inRows,
                [{number: 9, __id: myRowId2}, {number: -4, __id: myRowId1}, {number: -15, __id: myRowId3}]
            );
            
        }
  
    }
})
.addBatch({
        
    '(number.int) delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})

// Number tests (dec places)
.addBatch({
    
    '(number.dec) try to define a table with an invalid type': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':         {type: 'number'} // Missing min, max, decPlaces
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    '(number.dec) try to define a table with an invalid type': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':         {type: 'number', min: -50, max: 50} // Missing decPlaces
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    '(number.dec) try to define another table with an invalid type': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'fieldname':         {type: 'number', min: -50, max: 50, decPlaces: '1'} // Wrong type
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.notEqual(inError, null); // Not strict
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
            assert.strictEqual(inError.fieldName, 'fieldname');
        }
  
    }
})
.addBatch({
    
    '(number.dec) try to define valid table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'myField':         {type: 'number', min: -100, max: 100, decPlaces: 3}
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            assert.equal(inError, null);
        }
  
    }
})
.addBatch({
    
    '(number.dec) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': 'hi'
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(number.dec) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': -101
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(number.dec) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': 101
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(number.dec) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': 0.1111 // Too precise
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(number.dec) insert a valid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': 10.123
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowId) {
        
            assert.equal(inError, null);
            myRowId1 = inRowId;
        }
  
    }
})
.addBatch({
    
    '(number.dec) insert two more valid values': {
    
        'topic': function () {
            myTabler.insertMulti(
                'myTabler',
                [
                    {
                        'myField': 9.2
                    },
                    {
                        'myField': -15.33
                    }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowIds) {
        
            assert.equal(inError, null);
            myRowId2 = inRowIds[0];
            myRowId3 = inRowIds[1];
        }
  
    }
})
.addBatch({
    
    '(number.dec) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            // Equality testing not great for floats
            assert.deepEqual(
                inRows,
                [{myField: 10.123, __id: myRowId1}, {myField: 9.2, __id: myRowId2}, {myField: -15.33, __id: myRowId3}]
            );
            
        }
  
    }
})
.addBatch({
    
    '(number.dec) load the values with opposite ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'asc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(
                inRows,
                [{myField: -15.33, __id: myRowId3}, {myField: 9.2, __id: myRowId2}, {myField: 10.123, __id: myRowId1}]
            );
            
        }
  
    }
})
.addBatch({
    
    '(number.dec) update a value': {
    
        'topic': function () {
            myTabler.updateOne(
                'myTabler',
                {
                    'myField': -4.22
                },
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
  
    }
})
.addBatch({
    
    '(number.dec) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(
                inRows,
                [{myField: 9.2, __id: myRowId2}, {myField: -4.22, __id: myRowId1}, {myField: -15.33, __id: myRowId3}]
            );
            
        }
  
    }
})
.addBatch({
        
    '(number.dec) delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})

// Datetime tests
.addBatch({
    
    '(datetime) try to define valid table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'myField':         {type: 'datetime'}
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            assert.equal(inError, null);
        }
  
    }
})
.addBatch({
    
    '(datetime) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': 'hi'
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(datetime) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': -101
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(datetime) insert a valid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': (myCurDate = new Date())
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowId) {
        
            assert.equal(inError, null);
            myRowId1 = inRowId;
        }
  
    }
})
.addBatch({
    
    '(datetime) insert two more valid values': {
    
        'topic': function () {
            myTabler.insertMulti(
                'myTabler',
                [
                    {
                        'myField': new Date(0)
                    },
                    {
                        'myField': new Date(2015, 6, 12)
                    }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowIds) {
        
            assert.equal(inError, null);
            myRowId2 = inRowIds[0];
            myRowId3 = inRowIds[1];
        }
  
    }
})
.addBatch({
    
    '(datetime) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRows.length, 3)
            assert.strictEqual(inRows[0].__id, myRowId3);
            assert.strictEqual(inRows[1].__id, myRowId1);
            assert.strictEqual(inRows[2].__id, myRowId2);
            assert.strictEqual(inRows[0].myField.toString(), new Date(2015, 6, 12).toString());
            assert.strictEqual(inRows[1].myField.toString(), myCurDate.toString());
            assert.strictEqual(inRows[2].myField.toString(), new Date(0).toString());
            
        }
  
    }
})
.addBatch({
    
    '(datetime) load the values with opposite ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'asc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRows.length, 3)
            assert.strictEqual(inRows[0].__id, myRowId2);
            assert.strictEqual(inRows[1].__id, myRowId1);
            assert.strictEqual(inRows[2].__id, myRowId3);
            assert.strictEqual(inRows[0].myField.toString(), new Date(0).toString());
            assert.strictEqual(inRows[1].myField.toString(), myCurDate.toString());
            assert.strictEqual(inRows[2].myField.toString(), new Date(2015, 6, 12).toString());
            
        }
  
    }
})
.addBatch({
    
    '(datetime) update a value': {
    
        'topic': function () {
            myTabler.updateOne(
                'myTabler',
                {
                    'myField': new Date(10000)
                },
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
  
    }
})
.addBatch({
    
    '(datetime) load the values with ordering': {
    
        'topic': function () {
            myTabler.selectQuery(
                'myTabler',
                null, // All
                [], // All
                {
                    orderBy: 'myField',
                    orderDir: 'desc'
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.strictEqual(inRows.length, 3)
            assert.strictEqual(inRows[0].__id, myRowId3);
            assert.strictEqual(inRows[1].__id, myRowId1);
            assert.strictEqual(inRows[2].__id, myRowId2);
            assert.strictEqual(inRows[0].myField.toString(), new Date(2015, 6, 12).toString());
            assert.strictEqual(inRows[1].myField.toString(), new Date(10000).toString());
            assert.strictEqual(inRows[2].myField.toString(), new Date(0).toString());
            
        }
  
    }
})
.addBatch({
        
    '(datetime) delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})

// Object tests
.addBatch({
    
    '(object) try to define an invalid table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'myField':         {type: 'object'} // Missing byteLength
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_FIELD_DEFINITION');
        }
  
    }
})
.addBatch({
    
    '(object) try to define valid table': {
    
        'topic': function () {
            myTabler.defineTable(
                'myTabler',
                {
                    'myField':         {type: 'object', byteLength: 1000, schema: { properties: { a: { required: true }, b: { required: true } } } }
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
            assert.equal(inError, null);
        }
  
    }
})
.addBatch({
    
    '(object) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': function () {} // Not json serializable
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(object) insert an invalid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': {a: 1, h: 1} // Violates schema
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(object) insert an invalid value': {
    
        'topic': function () {
        
            var ins = {a: 1, b: 2};
            for (var i = 0; i < 1100; i++) {
                ins['_' + i] = i;
            }
        
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': ins
                },
                this.callback
            );
        },
        
        'and it fails': function (inError) {
            assert.isObject(inError);
            assert.strictEqual(inError.code, 'INVALID_VALUE');
            assert.strictEqual(inError.fieldName, 'myField');
        }
  
    }
})
.addBatch({
    
    '(object) insert a valid value': {
    
        'topic': function () {
            myTabler.insertOne(
                'myTabler',
                {
                    'myField': {a: 1, b: 2}
                },
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowId) {
            assert.equal(inError, null);
            myRowId1 = inRowId;
        }
  
    }
})
.addBatch({
    
    '(object) insert two more valid values': {
    
        'topic': function () {
            myTabler.insertMulti(
                'myTabler',
                [
                    {
                        'myField': {a: 1, b: 3}
                    },
                    {
                        'myField': {a: 2, b: 2}
                    }
                ],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRowIds) {
        
            assert.equal(inError, null);
            myRowId2 = inRowIds[0];
            myRowId3 = inRowIds[1];
        }
  
    }
})
.addBatch({
    
    '(object) load the values': {
    
        'topic': function () {
            myTabler.selectMulti(
                'myTabler',
                null, // All fields
                [myRowId1, myRowId2, myRowId3],
                this.callback
            );
        },
        
        'and it succeeds': function (inError, inRows) {
        
            assert.equal(inError, null);
            assert.deepEqual(inRows[myRowId1].myField, {a: 1, b: 2});
            assert.deepEqual(inRows[myRowId2].myField, {a: 1, b: 3});
            assert.deepEqual(inRows[myRowId3].myField, {a: 2, b: 2});
            
        }
  
    }
})
.addBatch({
    
    '(object) update a value': {
    
        'topic': function () {
            myTabler.updateOne(
                'myTabler',
                {
                    'myField': {a: 55, b: 2}
                },
                myRowId1,
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(inError, null);
            
        }
  
    }
})
.addBatch({
        
    '(object) delete the table': {
    
        'topic': function () {
            myTabler.deleteTable(
                'myTabler',
                this.callback
            );
        },
        
        'and it succeeds': function (inError) {
        
            assert.equal(null, inError);
                        
        }
        
    }

})



.run({
    error: false // Do not auto-detect the first argument of vows as an error, I check these myself
},
function () {

    if (process.argv[2] === 'jsonfile') {
        fs.unlinkSync('test.json');
    }
    
}
);


console.log(
    '-----\nRemember that if test fail then they might leave data, in which case if you run '
    + 'them again they could fail. Always try twice.\n-----'
);





